import os
import logging

supported_providers = ["travis", "appveyor"]


def get_continuous_integration_parameters():
    """
    Return a dictionary with information on a continuous integration build
    inferred from environment variables.

    The following dictionary keys are set (all values are also strings):
    - provider: name of CI provider, such as "travis" or "appveyor".
    - repo_slug: owner/name of the source code repository, i.e. "manubot/rootstock"
    - repo_owner: owner from repo_slug, i.e. "manubot"
    - repo_name: name from repo_slug, i.e. "rootstock"
    - commit: git commit being evaluated by the CI build
    - triggering_commit: git commit that triggered the CI build. For pull requests,
        CI services often build a merge commit with the default branch rather than
        the commit that was added to the pull request.
    - build_url: URL for the webpage with build details
    - job_url: URL for the webpage with job details
    """
    if os.getenv("CI", "false").lower() != "true":
        # No continuous integration environment detected
        return None

    if os.getenv("TRAVIS", "false") == "true":
        # https://docs.travis-ci.com/user/environment-variables/
        repo_slug = os.environ["TRAVIS_REPO_SLUG"]
        repo_owner, repo_name = repo_slug.split("/")
        return {
            "provider": "travis",
            "repo_slug": repo_slug,
            "repo_owner": repo_owner,
            "repo_name": repo_name,
            "commit": os.environ["TRAVIS_COMMIT"],
            "triggering_commit": os.getenv("TRAVIS_PULL_REQUEST_SHA")
            or os.environ["TRAVIS_COMMIT"],
            "build_url": os.environ["TRAVIS_BUILD_WEB_URL"],
            "job_url": os.environ["TRAVIS_JOB_WEB_URL"],
        }

    if os.getenv("APPVEYOR", "false").lower() == "true":
        # https://www.appveyor.com/docs/environment-variables/
        repo_slug = os.environ["APPVEYOR_REPO_NAME"]
        repo_owner, repo_name = repo_slug.split("/")
        provider_url = "{APPVEYOR_URL}/project/{APPVEYOR_ACCOUNT_NAME}/{APPVEYOR_PROJECT_SLUG}".format(
            **os.environ
        )
        build_url = f"{provider_url}/builds/{os.environ['APPVEYOR_BUILD_ID']}"
        return {
            "provider": "appveyor",
            "provider_account": os.environ["APPVEYOR_ACCOUNT_NAME"],
            "repo_slug": repo_slug,
            "repo_owner": repo_owner,
            "repo_name": repo_name,
            "commit": os.environ["APPVEYOR_REPO_COMMIT"],
            "triggering_commit": os.getenv("APPVEYOR_PULL_REQUEST_HEAD_COMMIT")
            or os.environ["APPVEYOR_REPO_COMMIT"],
            "build_url": build_url,
            "job_url": f"{provider_url}/build/job/{os.environ['APPVEYOR_JOB_ID']}",
            "artifact_url": f"{build_url}/artifacts",
        }

    logging.warning(
        "Detected CI environment variable, but get_continuous_integration_parameters "
        "did not detect environment variables for a supported CI provider. "
        "Supported providers are: {}".format(", ".join(supported_providers))
    )
    return None



import json
import logging
import urllib.request

import requests

from manubot.cite.pubmed import get_pubmed_ids_for_doi
from manubot.util import get_manubot_user_agent


def expand_short_doi(short_doi):
    """
    Convert a shortDOI to a regular DOI.
    """
    if not short_doi.startswith("10/"):
        raise ValueError(
            f"shortDOIs start with `10/`, but expand_short_doi received: {short_doi}"
        )
    url = f"https://doi.org/api/handles/{short_doi.lower()}"
    params = {"type": "HS_ALIAS"}
    response = requests.get(url, params=params)
    # response documentation at https://www.handle.net/proxy_servlet.html
    results = response.json()
    response_code = results.get("responseCode")  # Handle protocol response code
    if response_code == 100:
        raise ValueError(f"Handle not found. Double check short_doi: {short_doi}")
    if response_code == 200:
        raise ValueError(
            f"HS_ALIAS values not found. Double check short_doi: {short_doi}"
        )
    if response_code != 1:
        raise ValueError(
            f"Error response code of {response_code} returned by {response.url}"
        )
    values = results.get("values", [])
    for value in values:
        if value.get("type") == "HS_ALIAS":
            doi = value["data"]["value"]
            return doi.lower()
    raise RuntimeError(
        f'HS_ALIAS value not found by expand_short_doi("{short_doi}")\n'
        f"The following JSON was retrieved from {response.url}:\n"
        + json.dumps(results, indent=2)
    )


def get_short_doi_url(doi):
    """
    Get the shortDOI URL for a DOI.
    """
    quoted_doi = urllib.request.quote(doi)
    url = "http://shortdoi.org/{}?format=json".format(quoted_doi)
    headers = {"User-Agent": get_manubot_user_agent()}
    try:
        response = requests.get(url, headers=headers).json()
        short_doi = response["ShortDOI"]
        short_url = "https://doi.org/" + short_doi[3:]  # Remove "10/" prefix
        return short_url
    except Exception:
        logging.warning(f"shortDOI lookup failed for {doi}", exc_info=True)
        return None


def get_doi_csl_item_crosscite(doi):
    """
    Use Content Negotioation (https://crosscite.org/docs.html) to
    retrieve the CSL Item metadata for a DOI.
    """
    url = "https://doi.org/" + urllib.request.quote(doi)
    header = {
        "Accept": "application/vnd.citationstyles.csl+json",
        "User-Agent": get_manubot_user_agent(),
    }
    response = requests.get(url, headers=header)
    try:
        return response.json()
    except Exception as error:
        logging.error(
            f"Error fetching metadata for doi:{doi}.\n"
            f"Invalid response from {response.url}:\n{response.text}"
        )
        raise error


def get_doi_csl_item_zotero(doi):
    """
    Generate CSL JSON Data for a DOI using Zotero's translation-server.
    """
    from manubot.cite.zotero import get_csl_item

    return get_csl_item(f"doi:{doi}")


def augment_get_doi_csl_item(function):
    """
    Decorator providing edits to the csl_item returned by a get_doi_csl_item_* function.
    """

    def wrapper(doi: str):
        doi = doi.lower()
        csl_item = function(doi)
        csl_item["DOI"] = doi
        csl_item["URL"] = f"https://doi.org/{doi}"
        short_doi_url = get_short_doi_url(doi)
        if short_doi_url:
            csl_item["URL"] = short_doi_url
        try:
            csl_item.update(get_pubmed_ids_for_doi(doi))
        except Exception:
            logging.warning(
                f"Error calling get_pubmed_ids_for_doi for {doi}", exc_info=True
            )
        return csl_item

    return wrapper


@augment_get_doi_csl_item
def get_doi_csl_item(doi):
    """
    Generate CSL JSON Data for an DOI.

    This function uses a list of CSL JSON Item metadata retrievers, specified
    by the module-level variable `doi_retrievers`. The methods are attempted
    in order, with this function returning the metadata from the first
    non-failing method.
    """
    # FIXME: this function is repetitive with other get_*_csl_item functions.
    for retriever in doi_retrievers:
        try:
            return retriever(doi)
        except Exception as error:
            logging.warning(
                f"Error in {retriever.__name__} for {doi} "
                f"due to a {error.__class__.__name__}:\n{error}"
            )
            logging.info(error, exc_info=True)
    raise Exception(f"all get_doi_csl_item methods failed for {doi}")


doi_retrievers = [get_doi_csl_item_crosscite, get_doi_csl_item_zotero]



import logging


def cli_process(args):
    args_dict = vars(args)

    # Set paths for content
    content_dir = args.content_directory
    if not content_dir.is_dir():
        logging.warning(f"content directory does not exist: {content_dir}")
    args_dict["citation_tags_path"] = content_dir.joinpath("citation-tags.tsv")
    args_dict["meta_yaml_path"] = content_dir.joinpath("metadata.yaml")
    args_dict["manual_references_paths"] = sorted(
        content_dir.rglob("manual-references*.*")
    )

    # Set paths for output
    output_dir = args.output_directory
    output_dir.mkdir(parents=True, exist_ok=True)
    args_dict["manuscript_path"] = output_dir.joinpath("manuscript.md")
    args_dict["citations_path"] = output_dir.joinpath("citations.tsv")
    args_dict["references_path"] = output_dir.joinpath("references.json")
    args_dict["variables_path"] = output_dir.joinpath("variables.json")

    # Set paths for caching
    args_dict["cache_directory"] = args.cache_directory or output_dir
    args.cache_directory.mkdir(parents=True, exist_ok=True)
    args_dict["requests_cache_path"] = str(
        args.cache_directory.joinpath("requests-cache")
    )

    from manubot.process.util import prepare_manuscript

    prepare_manuscript(args)



import json
import logging
import pathlib

from manubot import __version__ as manubot_version
from manubot.cite.citekey import shorten_citekey


def load_bibliography(path) -> list:
    """
    Load a bibliography as CSL Items (a CSL JSON Python object).
    For paths that already contain CSL Items (inferred from a .json or .yaml extension),
    parse these files directly. Otherwise, delegate conversion to CSL Items to pandoc-citeproc.
    """
    path = pathlib.Path(path)
    use_pandoc_citeproc = True
    try:
        if path.suffix == ".json":
            use_pandoc_citeproc = False
            with path.open(encoding="utf-8-sig") as read_file:
                csl_items = json.load(read_file)
        if path.suffix == ".yaml":
            use_pandoc_citeproc = False
            import yaml

            with path.open(encoding="utf-8-sig") as read_file:
                csl_items = yaml.safe_load(read_file)
    except Exception:
        logging.exception(f"process.load_bibliography: error parsing {path}.\n")
        csl_items = []
    if use_pandoc_citeproc:
        from manubot.pandoc.bibliography import (
            load_bibliography as load_bibliography_pandoc,
        )

        csl_items = load_bibliography_pandoc(path)
    if not isinstance(csl_items, list):
        logging.error(
            f"process.load_bibliography: csl_items read from {path} are of type {type(csl_items)}. "
            "Setting csl_items to an empty list."
        )
        csl_items = []
    from manubot.cite.csl_item import CSL_Item

    csl_items = [CSL_Item(csl_item) for csl_item in csl_items]
    return csl_items


def load_manual_references(paths=[], extra_csl_items=[]) -> dict:
    """
    Read manual references (overrides) from files specified by a list of paths.
    Returns a standard_citation to CSL Item dictionary. extra_csl_items specifies
    JSON CSL stored as a Python object, to be used in addition to the CSL JSON
    stored as text in the file specified by path. Set paths=[] to only use extra_csl_items.
    """
    from manubot.cite.csl_item import CSL_Item

    csl_items = []
    for path in paths:
        path = pathlib.Path(path)
        if not path.is_file():
            logging.warning(
                f"process.load_bibliographies is skipping a non-existent path: {path}"
            )
            continue
        for csl_item in load_bibliography(path):
            csl_item.note_append_text(
                f"This CSL JSON Item was loaded by Manubot v{manubot_version} from a manual reference file."
            )
            csl_item.note_append_dict({"manual_reference_filename": path.name})
            csl_items.append(csl_item)
    csl_items.extend(map(CSL_Item, extra_csl_items))
    manual_refs = dict()
    for csl_item in csl_items:
        try:
            csl_item.standardize_id()
        except Exception:
            csl_item_str = json.dumps(csl_item, indent=2)
            logging.info(
                f"Skipping csl_item where setting standard_id failed:\n{csl_item_str}",
                exc_info=True,
            )
            continue
        standard_id = csl_item["id"]
        csl_item.set_id(shorten_citekey(standard_id))
        csl_item.clean()
        manual_refs[standard_id] = csl_item
    return manual_refs



import pytest

from manubot.cite.isbn import (
    get_isbn_csl_item_citoid,
    get_isbn_csl_item_isbnlib,
    get_isbn_csl_item_zotero,
)


@pytest.mark.xfail(reason="Quotation in title removed at some upstream point")
def test_citekey_to_csl_item_isbnlib_title_with_quotation_mark():
    csl_item = get_isbn_csl_item_isbnlib("9780312353780")
    assert csl_item["type"] == "book"
    assert csl_item["title"].startswith('"F" is for Fugitive')


def test_get_isbn_csl_item_citoid_weird_date():
    """
    isbn:9780719561023 has a date value of "(2004 printing)"
    https://en.wikipedia.org/api/rest_v1/data/citation/mediawiki/9780719561023
    """
    csl_item = get_isbn_csl_item_citoid("9780719561023")
    assert csl_item["issued"]["date-parts"] == [[2004]]
    assert csl_item["ISBN"] == "9780719561023"


def test_get_isbn_csl_item_citoid_not_found():
    """
    isbn:9781439566039 is not found by Citoid:
    https://en.wikipedia.org/api/rest_v1/data/citation/mediawiki/9781439566039
    """
    with pytest.raises(KeyError, match=r"Metadata for ISBN [0-9]{10,13} not found"):
        get_isbn_csl_item_citoid("9781439566039")


def test_get_isbn_csl_item_zotero_with_note_issue():
    """
    translation-server returns two metadata records for this ISBN.
    The second has itemType=note and previously caused CSL export to fail.
    https://github.com/zotero/translation-server/issues/67
    """
    isbn = "9780262517638"
    csl_item = get_isbn_csl_item_zotero(isbn)
    assert csl_item["author"][0]["family"] == "Suber"



from manubot.process.manuscript import get_citekeys, update_manuscript_citekeys


def test_get_citekeys_1():
    text = """
    Sci-Hub has released article request records from its server logs, covering 165 days from September 2015 through February 2016 [@doi:10.1126/science.352.6285.508; @doi:10.1126/science.aaf5664; @doi:10.5061/dryad.q447c/1].
    We filtered for valid requests by excluding DOIs not included in our literature catalog and omitting requests that occurred before an article's publication date.

    Figure {@fig:citescore}B shows that articles from highly cited journals were visited much more frequently on average.

    @10.5061/bad_doi says blah but @url:https://www.courtlistener.com/docket/4355308/1/elsevier-inc-v-sci-hub/ disagrees.
    """
    citekeys = get_citekeys(text)
    expected = sorted(
        [
            "doi:10.1126/science.352.6285.508",
            "doi:10.1126/science.aaf5664",
            "doi:10.5061/dryad.q447c/1",
            "url:https://www.courtlistener.com/docket/4355308/1/elsevier-inc-v-sci-hub/",
        ]
    )
    assert citekeys == expected


def test_update_manuscript_citekeys():
    """
    Test that text does not get converted to:

    > our new Manubot tool [@cTN2TQIL-rootstock; @cTN2TQIL] for automating
    manuscript generation.

    See https://github.com/manubot/manubot/issues/9
    """
    string_to_id = {
        "url:https://github.com/manubot/manubot": "mNMayr3f",
        "url:https://github.com/greenelab/manubot-rootstock": "1B7Y2HVtw",
    }
    text = "our new Manubot tool [@url:https://github.com/greenelab/manubot-rootstock; @url:https://github.com/manubot/manubot] for automating manuscript generation."
    text = update_manuscript_citekeys(text, string_to_id)
    assert "[@1B7Y2HVtw; @mNMayr3f]" in text



import os
import re

import pytest

from ..ci import get_continuous_integration_parameters


@pytest.mark.skipif(
    "TRAVIS" not in os.environ,
    reason="tests environment variables set by Travis builds only",
)
@pytest.mark.skipif(
    os.getenv("TRAVIS_REPO_SLUG") != "manubot/manubot", reason="test fails on forks"
)
def test_get_continuous_integration_parameters_travis():
    info = get_continuous_integration_parameters()
    assert info is not None
    assert info["provider"] == "travis"
    assert info["repo_slug"] == "manubot/manubot"
    assert info["repo_owner"] == "manubot"
    assert info["repo_name"] == "manubot"
    assert info["commit"]
    assert info["triggering_commit"]
    assert info["build_url"].startswith("https://travis-ci.com/manubot/manubot/builds/")
    assert info["job_url"].startswith("https://travis-ci.com/manubot/manubot/jobs/")


@pytest.mark.skipif(
    "APPVEYOR" not in os.environ,
    reason="tests environment variables set by AppVeyor builds only",
)
@pytest.mark.skipif(
    os.getenv("APPVEYOR_REPO_NAME") != "manubot/manubot", reason="test fails on forks"
)
def test_get_continuous_integration_parameters_appveyor():
    info = get_continuous_integration_parameters()
    assert info is not None
    assert info["provider"] == "appveyor"
    assert info["provider_account"] == "manubot"
    assert info["repo_slug"] == "manubot/manubot"
    assert info["repo_owner"] == "manubot"
    assert info["repo_name"] == "manubot"
    assert info["commit"]
    assert info["triggering_commit"]
    assert info["build_url"].startswith(
        "https://ci.appveyor.com/project/manubot/manubot/builds/"
    )
    assert info["job_url"].startswith(
        "https://ci.appveyor.com/project/manubot/manubot/build/job/"
    )
    assert re.fullmatch(
        pattern=r"https://ci\.appveyor\.com/project/manubot/manubot/builds/[0-9]+/artifacts",
        string=info["artifact_url"],
    )



import collections
import logging
import re
import xml.etree.ElementTree

import requests

from manubot.util import get_manubot_user_agent


def get_arxiv_csl_item(arxiv_id):
    """
    Return csl_item item for an arXiv record.

    arxiv_id can be versioned, like `1512.00567v2`, or versionless, like
    `1512.00567`. If versionless, the arXiv API will return metadata for the
    latest version. Legacy IDs, such as `cond-mat/0703470v2`, are also
    supported.

    If arXiv has an associated DOI for the record, a warning is logged to
    alert the user that an alternative version of record exists.

    References:
    - https://arxiv.org/help/api/index
    - http://citeproc-js.readthedocs.io/en/latest/csl-json/markup.html
    - https://github.com/citation-style-language/schema/blob/master/csl-data.json
    """
    url = "https://export.arxiv.org/api/query"
    params = {"id_list": arxiv_id, "max_results": 1}
    headers = {"User-Agent": get_manubot_user_agent()}
    response = requests.get(url, params, headers=headers)

    # XML namespace prefixes
    prefix = "{http://www.w3.org/2005/Atom}"
    alt_prefix = "{http://arxiv.org/schemas/atom}"

    # Parse XML
    xml_tree = xml.etree.ElementTree.fromstring(response.text)
    (entry,) = xml_tree.findall(prefix + "entry")

    # Create dictionary for CSL Item
    csl_item = collections.OrderedDict()

    # Extract versioned arXiv ID
    url = entry.findtext(prefix + "id")
    pattern = re.compile(r"arxiv.org/abs/(.+)")
    match = pattern.search(url)
    versioned_id = match.group(1)
    csl_item["number"] = versioned_id
    _, csl_item["version"] = versioned_id.rsplit("v", 1)
    csl_item["URL"] = "https://arxiv.org/abs/" + versioned_id

    # Extrat CSL title field
    csl_item["title"] = entry.findtext(prefix + "title")

    # Extract CSL date field
    published = entry.findtext(prefix + "published")
    published, _ = published.split("T", 1)
    csl_item["issued"] = {"date-parts": [[int(x) for x in published.split("-")]]}

    # Extract authors
    authors = list()
    for elem in entry.findall(prefix + "author"):
        name = elem.findtext(prefix + "name")
        author = {"literal": name}
        authors.append(author)
    csl_item["author"] = authors

    # Set publisher to arXiv
    csl_item["container-title"] = "arXiv"
    csl_item["publisher"] = "arXiv"

    # Extract abstract
    abstract = entry.findtext(prefix + "summary").strip()
    if abstract:
        csl_item["abstract"] = abstract

    # Check if the article has been published with a DOI
    DOI = entry.findtext("{http://arxiv.org/schemas/atom}doi")
    if DOI:
        csl_item["DOI"] = DOI
        journal_ref = entry.findtext(alt_prefix + "journal_ref")
        msg = f"arXiv article {arxiv_id} published at https://doi.org/{DOI}"
        if journal_ref:
            msg += f" — {journal_ref}"
        logging.warning(msg)
    # Set CSL type to report for preprint
    csl_item["type"] = "report"
    return csl_item



import json
import pathlib
import shutil
import subprocess

import pytest

from manubot.util import shlex_join
from manubot.pandoc.util import get_pandoc_version


def test_cite_command_empty():
    process = subprocess.run(
        ["manubot", "cite"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    print(process.stderr)
    assert process.returncode == 2
    assert "the following arguments are required: citekeys" in process.stderr


def test_cite_command_stdout():
    process = subprocess.run(
        ["manubot", "cite", "arxiv:1806.05726v1"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    print(process.stderr)
    assert process.returncode == 0
    (csl,) = json.loads(process.stdout)
    assert csl["URL"] == "https://arxiv.org/abs/1806.05726v1"


def test_cite_command_file(tmpdir):
    path = pathlib.Path(tmpdir) / "csl-items.json"
    process = subprocess.run(
        ["manubot", "cite", "--output", str(path), "arxiv:1806.05726v1"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    print(process.stderr.decode())
    assert process.returncode == 0
    with path.open() as read_file:
        (csl,) = json.load(read_file)
    assert csl["URL"] == "https://arxiv.org/abs/1806.05726v1"


pandoc_version = get_pandoc_version()


@pytest.mark.skipif(
    not shutil.which("pandoc"), reason="pandoc installation not found on system"
)
@pytest.mark.skipif(
    not shutil.which("pandoc-citeproc"),
    reason="pandoc-citeproc installation not found on system",
)
class Base_cite_command_render_stdout:
    """
    Expecting reference values for test to be at files on path:
    cite-command-rendered/references-{format}-{pandoc_stamp}.{extension}

    Examples:
    - references-html-2.7.2.html  
    - references-plain-2.7.2.txt
    - references-jats-2.7.2.xml
    - references-jats-2.7.3.xml
    - references-markdown-2.7.2.md

    2.7.2 was the current pandoc version on CI builds for these builds,
    but a newer version 2.7.3 was available too and giving a different output
    for xml (jats) output.

    Filenames must be adjusted accodingly when current pandoc version changes.

    Run tests locally skipping this test suit (makes sense if your local pandoc
    version is different from pandoc version used by Travis and Appveyor):

        pytest -v -m "not pandoc_version_sensitive"
        
    See .travis.yml and .appveyor.yml to find out current pandoc version used 
    for testing.
    """

    pandoc_stamp = ".".join(map(str, pandoc_version))

    @classmethod
    def expected_output(cls, format, extension):
        filename = f"references-{format}-{cls.pandoc_stamp}.{extension}"
        return (
            pathlib.Path(__file__)
            .parent.joinpath("cite-command-rendered", filename)
            .read_text()
        )
        # Note: We wanted to introduce encoding='utf-8-sig' above
        # for compatability, but that makes tests fail on Travis for
        # Windows default encoding.

    @staticmethod
    def render(format_args):
        args = [
            "manubot",
            "cite",
            "--render",
            "--csl",
            "https://github.com/greenelab/manubot-rootstock/raw/e83e51dcd89256403bb787c3d9a46e4ee8d04a9e/build/assets/style.csl",
            "arxiv:1806.05726v1",
            "doi:10.7717/peerj.338",
            "pmid:29618526",
        ] + format_args
        process = subprocess.run(
            args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        print(shlex_join(process.args))
        print(process.stdout)
        print(process.stderr)
        return process.stdout


@pytest.mark.pandoc_version_sensitive
@pytest.mark.skipif(
    pandoc_version < (2, 0),
    reason="Test requires pandoc >= 2.0 to support --lua-filter and --csl=URL",
)
class Test_cite_command_render_stdout_above_pandoc_v2(Base_cite_command_render_stdout):
    def test_no_arg(self):
        assert self.render([]) == self.expected_output("plain", "txt")

    def test_plain(self):
        assert self.render(["--format", "plain"]) == self.expected_output(
            "plain", "txt"
        )


@pytest.mark.pandoc_version_sensitive
@pytest.mark.skipif(
    pandoc_version < (2, 5),
    reason=("Testing markdown, html or jats formats " "assumes pandoc >= 2.5"),
)
class Test_cite_command_render_stdout_above_pandoc_v2_5(
    Base_cite_command_render_stdout
):
    def test_markdown(self):
        assert self.render(["--format", "markdown"]) == self.expected_output(
            "markdown", "md"
        )

    def test_html(self):
        assert self.render(["--format", "html"]) == self.expected_output("html", "html")

    def test_jats(self):
        assert self.render(["--format", "jats"]) == self.expected_output("jats", "xml")


def teardown_module(module):
    """
    Avoid too many requests to NCBI E-Utils in the test_pubmed.py,
    which is executed following this module. E-Utility requests are
    capped at 3 per second, which is usually controled by _get_eutils_rate_limiter,
    but this does not seem to work across test modules.
    """
    import time

    time.sleep(1)



from manubot.cite.wikidata import get_wikidata_csl_item


def test_get_wikidata_csl_item():
    """
    Test metadata extraction from https://www.wikidata.org/wiki/Q50051684
    """
    wikidata_id = "Q50051684"
    csl_item = get_wikidata_csl_item(wikidata_id)
    assert (
        "Sci-Hub provides access to nearly all scholarly literature"
        in csl_item["title"]
    )
    assert csl_item["container-title"] == "eLife"
    assert csl_item["DOI"] == "10.7554/elife.32822"


def test_get_wikidata_csl_item_author_ordering():
    """
    Test extraction of author ordering from https://www.wikidata.org/wiki/Q50051684.
    Wikidata uses a "series ordinal" qualifier that must be considered or else author
    ordering may be wrong.

    Author ordering was previously not properly set by the Wikidata translator
    https://github.com/zotero/translators/issues/1790
    """
    wikidata_id = "Q50051684"
    csl_item = get_wikidata_csl_item(wikidata_id)
    family_names = [author["family"] for author in csl_item["author"]]
    print(family_names)
    assert family_names == [
        "Himmelstein",
        "Romero",
        "Levernier",
        "Munro",
        "McLaughlin",
        "Greshake",  # actually should be Greshake Tzovaras
        "Greene",
    ]



"""Tests rest of functions in manubot.cite, not covered by test_citekey_api.py."""

import pytest

from manubot.cite.citekey import (
    citekey_pattern,
    shorten_citekey,
    infer_citekey_prefix,
    inspect_citekey,
)


@pytest.mark.parametrize(
    "citation_string",
    [
        ("@doi:10.5061/dryad.q447c/1"),
        ("@arxiv:1407.3561v1"),
        ("@doi:10.1007/978-94-015-6859-3_4"),
        ("@tag:tag_with_underscores"),
        ("@tag:tag-with-hyphens"),
        ("@url:https://greenelab.github.io/manubot-rootstock/"),
        ("@tag:abc123"),
        ("@tag:123abc"),
    ],
)
def test_citekey_pattern_match(citation_string):
    match = citekey_pattern.fullmatch(citation_string)
    assert match


@pytest.mark.parametrize(
    "citation_string",
    [
        ("doi:10.5061/dryad.q447c/1"),
        ("@tag:abc123-"),
        ("@tag:abc123_"),
        ("@-tag:abc123"),
        ("@_tag:abc123"),
    ],
)
def test_citekey_pattern_no_match(citation_string):
    match = citekey_pattern.fullmatch(citation_string)
    assert match is None


@pytest.mark.parametrize(
    "standard_citekey,expected",
    [
        ("doi:10.5061/dryad.q447c/1", "kQFQ8EaO"),
        ("arxiv:1407.3561v1", "16kozZ9Ys"),
        ("pmid:24159271", "11sli93ov"),
        ("url:http://blog.dhimmel.com/irreproducible-timestamps/", "QBWMEuxW"),
    ],
)
def test_shorten_citekey(standard_citekey, expected):
    short_citekey = shorten_citekey(standard_citekey)
    assert short_citekey == expected


@pytest.mark.parametrize(
    "citekey",
    [
        "doi:10.7717/peerj.705",
        "doi:10/b6vnmd",
        "pmcid:PMC4304851",
        "pmid:25648772",
        "arxiv:1407.3561",
        "isbn:978-1-339-91988-1",
        "isbn:1-339-91988-5",
        "wikidata:Q1",
        "wikidata:Q50051684",
        "url:https://peerj.com/articles/705/",
    ],
)
def test_inspect_citekey_passes(citekey):
    """
    These citekeys should pass inspection by inspect_citekey.
    """
    assert inspect_citekey(citekey) is None


@pytest.mark.parametrize(
    ["citekey", "contains"],
    [
        ("doi:10.771/peerj.705", "Double check the DOI"),
        ("doi:10/b6v_nmd", "Double check the shortDOI"),
        ("doi:7717/peerj.705", "must start with '10.'"),
        ("doi:b6vnmd", "must start with '10.'"),
        ("pmcid:25648772", "must start with 'PMC'"),
        (
            "pmid:PMC4304851",
            "Should 'pmid:PMC4304851' switch the citation source to 'pmcid'?",
        ),
        ("isbn:1-339-91988-X", "identifier violates the ISBN syntax"),
        ("wikidata:P212", "item IDs must start with 'Q'"),
        ("wikidata:QABCD", "does not conform to the Wikidata regex"),
    ],
)
def test_inspect_citekey_fails(citekey, contains):
    """
    These citekeys should fail inspection by inspect_citekey.
    """
    report = inspect_citekey(citekey)
    assert report is not None
    assert isinstance(report, str)
    assert contains in report


@pytest.mark.parametrize(
    ["citation", "expect"],
    [
        ("doi:not-a-real-doi", "doi:not-a-real-doi"),
        ("DOI:not-a-real-doi", "doi:not-a-real-doi"),
        ("uRl:mixed-case-prefix", "url:mixed-case-prefix"),
        ("raw:raw-citation", "raw:raw-citation"),
        ("no-prefix", "raw:no-prefix"),
        ("no-prefix:but-colon", "raw:no-prefix:but-colon"),
    ],
)
def test_infer_citekey_prefix(citation, expect):
    assert infer_citekey_prefix(citation) == expect



import pathlib
import re
import shlex
import subprocess

import pytest

readme_path = pathlib.Path(__file__).parent.parent.parent / "README.md"
readme = readme_path.read_text(encoding="utf-8-sig")

template = r"""
<!-- test codeblock contains output of `{command}` -->
```
{output}```
"""
pattern = template.format(command=r"(?P<command>.+?)", output=r"(?P<output>.+?)")
pattern = re.compile(pattern, re.DOTALL)
matches = list(pattern.finditer(readme))


@pytest.mark.parametrize(
    argnames=["command", "expected"],
    argvalues=[match.groups() for match in matches],
    ids=[match.group("command") for match in matches],
)
def test_readme_codeblock_contains_output_from(command, expected):
    """
    If this test fails, ensure that codeblocks in README.md have the correct
    output. To enable this check for output in a codeblock, use the following
    construct:

    <!-- test codeblock contains output of `{command}` -->
    ```
    {expected}
    ```
    """
    output = _get_output_from(command)
    assert output == expected


def _get_output_from(command):
    return subprocess.check_output(shlex.split(command), universal_newlines=True)


def _match_to_repl(match):
    template_dict = match.groupdict()
    template_dict["output"] = _get_output_from(template_dict["command"])
    return template.format(**template_dict)


if __name__ == "__main__":
    """
    Run `python manubot/tests/test_readme.py` to populate README
    codeblocks with output from the specified commands.
    """
    repl_readme = pattern.sub(repl=_match_to_repl, string=readme)
    readme_path.write_text(repl_readme, encoding="utf-8")



import collections
import datetime
import json
import logging
import pathlib
import re

from manubot.cite.citekey import citekey_pattern, is_valid_citekey


def get_citekeys(text):
    """
    Extract the deduplicated list of citations in a text. Citations that are
    clearly invalid such as `doi:/453` are not returned.
    """
    citekeys = set(citekey_pattern.findall(text))
    citekeys = filter(
        lambda x: is_valid_citekey(
            x, allow_tag=True, allow_raw=True, allow_pandoc_xnos=True
        ),
        citekeys,
    )
    return sorted(citekeys)


def get_text(directory):
    """
    Return a concatenated string of section texts from the specified directory.
    Text files should be UTF-8 encoded.
    """
    section_dir = pathlib.Path(directory)
    paths = sorted(section_dir.glob("[0-9]*.md"))
    name_to_text = collections.OrderedDict()
    for path in paths:
        name_to_text[path.stem] = path.read_text(encoding="utf-8-sig")
    logging.info("Manuscript content parts:\n" + "\n".join(name_to_text))
    return "\n\n".join(name_to_text.values()) + "\n"


def update_manuscript_citekeys(text, old_to_new):
    """
    Replace citation keys according to the old_to_new dictionary.
    Useful for converting citation keys to shortened versions
    that are appropriate for pandoc.

    `text` is markdown source text

    `old_to_new` is a dictionary like:
    doi:10.7287/peerj.preprints.3100v1 → 11cb5HXoY
    """
    for old, new in old_to_new.items():
        text = re.sub(
            pattern=re.escape("@" + old) + r"(?![\w:.#$%&\-+?<>~/]*[a-zA-Z0-9/])",
            repl="@" + new,
            string=text,
        )
    return text


def get_manuscript_stats(text, citekeys_df):
    """
    Compute manuscript statistics.
    """
    stats = collections.OrderedDict()

    # Number of distinct references by type
    ref_counts = (
        citekeys_df.standard_citekey.drop_duplicates()
        .map(lambda x: x.split(":")[0])
        .pipe(collections.Counter)
    )
    ref_counts["total"] = sum(ref_counts.values())
    stats["reference_counts"] = ref_counts
    stats["word_count"] = len(text.split())
    logging.info(f"Generated manscript stats:\n{json.dumps(stats, indent=2)}")
    return stats


def datetime_now():
    """
    Return the current datetime, with timezone awareness
    https://stackoverflow.com/a/39079819/4651668
    """
    tzinfo = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo
    return datetime.datetime.now(tzinfo)



def get_wikidata_csl_item(identifier):
    """
    Get a CSL JSON item with the citation metadata for a Wikidata item.
    identifier should be a Wikidata item ID corresponding to a citeable
    work, such as Q50051684.
    """
    url = f"https://www.wikidata.org/wiki/{identifier}"
    from manubot.cite.url import get_url_csl_item_zotero

    csl_item = get_url_csl_item_zotero(url)
    if "DOI" in csl_item:
        csl_item["DOI"] = csl_item["DOI"].lower()
    if not "URL" in csl_item:
        csl_item["URL"] = url
    return csl_item



"""Tests API-level functions in manubot.cite. Both functions are found in citekey.py"""

import pytest

from manubot.cite import citekey_to_csl_item, standardize_citekey


@pytest.mark.parametrize(
    "citekey,expected",
    [
        ("doi:10.5061/DRYAD.q447c/1", "doi:10.5061/dryad.q447c/1"),
        ("doi:10.5061/dryad.q447c/1", "doi:10.5061/dryad.q447c/1"),
        ("doi:10/b6vnmd", "doi:10.1016/s0933-3657(96)00367-3"),
        ("doi:10/B6VNMD", "doi:10.1016/s0933-3657(96)00367-3"),
        (
            "doi:10/xxxxxxxxxxxxxYY",
            "doi:10/xxxxxxxxxxxxxyy",
        ),  # passthrough non-existent shortDOI
        ("pmid:24159271", "pmid:24159271"),
        ("isbn:1339919885", "isbn:9781339919881"),
        ("isbn:1-339-91988-5", "isbn:9781339919881"),
        ("isbn:978-0-387-95069-3", "isbn:9780387950693"),
        ("isbn:9780387950938", "isbn:9780387950938"),
        ("isbn:1-55860-510-X", "isbn:9781558605107"),
        ("isbn:1-55860-510-x", "isbn:9781558605107"),
    ],
)
def test_standardize_citekey(citekey, expected):
    """
    Standardize identifiers based on their source
    """
    output = standardize_citekey(citekey)
    assert output == expected


@pytest.mark.xfail(reason="https://twitter.com/dhimmel/status/950443969313419264")
def test_citekey_to_csl_item_doi_datacite():
    citekey = "doi:10.7287/peerj.preprints.3100v1"
    csl_item = citekey_to_csl_item(citekey)
    assert csl_item["id"] == "11cb5HXoY"
    assert csl_item["URL"] == "https://doi.org/10.7287/peerj.preprints.3100v1"
    assert csl_item["DOI"] == "10.7287/peerj.preprints.3100v1"
    assert csl_item["type"] == "report"
    assert (
        csl_item["title"]
        == "Sci-Hub provides access to nearly all scholarly literature"
    )
    authors = csl_item["author"]
    assert authors[0]["family"] == "Himmelstein"
    assert authors[-1]["family"] == "Greene"


def test_citekey_to_csl_item_arxiv():
    citekey = "arxiv:cond-mat/0703470v2"
    csl_item = citekey_to_csl_item(citekey)
    assert csl_item["id"] == "ES92tcdg"
    assert csl_item["URL"] == "https://arxiv.org/abs/cond-mat/0703470v2"
    assert csl_item["number"] == "cond-mat/0703470v2"
    assert csl_item["version"] == "2"
    assert csl_item["type"] == "report"
    assert csl_item["container-title"] == "arXiv"
    assert csl_item["title"] == "Portraits of Complex Networks"
    authors = csl_item["author"]
    assert authors[0]["literal"] == "J. P. Bagrow"
    assert csl_item["DOI"] == "10.1209/0295-5075/81/68004"


def test_citekey_to_csl_item_pmc():
    """
    https://api.ncbi.nlm.nih.gov/lit/ctxp/v1/pmc/?format=csl&id=3041534
    """
    citekey = f"pmcid:PMC3041534"
    csl_item = citekey_to_csl_item(citekey)
    assert csl_item["id"] == "RoOhUFKU"
    assert csl_item["URL"] == "https://www.ncbi.nlm.nih.gov/pmc/articles/PMC3041534/"
    assert csl_item["container-title-short"] == "Summit Transl Bioinform"
    assert (
        csl_item["title"]
        == "Secondary Use of EHR: Data Quality Issues and Informatics Opportunities"
    )
    authors = csl_item["author"]
    assert authors[0]["family"] == "Botsis"
    assert csl_item["PMID"] == "21347133"
    assert csl_item["PMCID"] == "PMC3041534"
    assert "generated by Manubot" in csl_item["note"]
    assert "standard_id: pmcid:PMC3041534" in csl_item["note"]


def test_citekey_to_csl_item_pubmed_1():
    """
    Generated from XML returned by
    https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=21347133&rettype=full
    """
    citekey = "pmid:21347133"
    csl_item = citekey_to_csl_item(citekey)
    assert csl_item["id"] == "y9ONtSZ9"
    assert csl_item["type"] == "article-journal"
    assert csl_item["URL"] == "https://www.ncbi.nlm.nih.gov/pubmed/21347133"
    assert csl_item["container-title"] == "Summit on translational bioinformatics"
    assert (
        csl_item["title"]
        == "Secondary Use of EHR: Data Quality Issues and Informatics Opportunities."
    )
    assert csl_item["issued"]["date-parts"] == [[2010, 3, 1]]
    authors = csl_item["author"]
    assert authors[0]["given"] == "Taxiarchis"
    assert authors[0]["family"] == "Botsis"
    assert csl_item["PMID"] == "21347133"
    assert csl_item["PMCID"] == "PMC3041534"


def test_citekey_to_csl_item_pubmed_2():
    """
    Generated from XML returned by
    https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=27094199&rettype=full
    """
    citekey = "pmid:27094199"
    csl_item = citekey_to_csl_item(citekey)
    print(csl_item)
    assert csl_item["id"] == "alaFV9OY"
    assert csl_item["type"] == "article-journal"
    assert csl_item["URL"] == "https://www.ncbi.nlm.nih.gov/pubmed/27094199"
    assert csl_item["container-title"] == "Circulation. Cardiovascular genetics"
    assert csl_item["container-title-short"] == "Circ Cardiovasc Genet"
    assert csl_item["page"] == "179-84"
    assert (
        csl_item["title"]
        == "Genetic Association-Guided Analysis of Gene Networks for the Study of Complex Traits."
    )
    assert csl_item["issued"]["date-parts"] == [[2016, 4]]
    authors = csl_item["author"]
    assert authors[0]["given"] == "Casey S"
    assert authors[0]["family"] == "Greene"
    assert csl_item["PMID"] == "27094199"
    assert csl_item["DOI"] == "10.1161/circgenetics.115.001181"


def test_citekey_to_csl_item_pubmed_with_numeric_month():
    """
    Generated from XML returned by
    https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=29028984&rettype=full

    See https://github.com/manubot/manubot/issues/69
    """
    citekey = "pmid:29028984"
    csl_item = citekey_to_csl_item(citekey)
    print(csl_item)
    assert csl_item["issued"]["date-parts"] == [[2018, 3, 15]]


def test_citekey_to_csl_item_pubmed_book():
    """
    Extracting CSL metadata from books in PubMed is not supported.
    Logic not implemented to parse XML returned by
    https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=29227604&rettype=full
    """
    with pytest.raises(NotImplementedError):
        citekey_to_csl_item("pmid:29227604")


def test_citekey_to_csl_item_isbn():
    csl_item = citekey_to_csl_item("isbn:9780387950693")
    assert csl_item["type"] == "book"
    assert csl_item["title"] == "Complex analysis"



import pytest

from manubot.cite.pubmed import (
    get_pmcid_and_pmid_for_doi,
    get_pmid_for_doi,
    get_pubmed_ids_for_doi,
)


@pytest.mark.parametrize(
    ("doi", "pmid"),
    [
        ("10.1098/rsif.2017.0387", "29618526"),  # in PubMed and PMC
        ("10.1161/CIRCGENETICS.115.001181", "27094199"),  # in PubMed but not PMC
        ("10.7717/peerj-cs.134", None),  # DOI in journal not indexed by PubMed
        ("10.1161/CIRC", None),  # invalid DOI
    ],
)
def test_get_pmid_for_doi(doi, pmid):
    output = get_pmid_for_doi(doi)
    assert pmid == output


@pytest.mark.parametrize(
    ("doi", "id_dict"),
    [
        ("10.1098/rsif.2017.0387", {"PMCID": "PMC5938574", "PMID": "29618526"}),
        ("10.7554/ELIFE.32822", {"PMCID": "PMC5832410", "PMID": "29424689"}),
        ("10.1161/CIRCGENETICS.115.001181", {}),  # only in PubMed, not in PMC
        ("10.7717/peerj.000", {}),  # Non-existent DOI
        ("10.peerj.000", {}),  # malformed DOI
    ],
)
def test_get_pmcid_and_pmid_for_doi(doi, id_dict):
    output = get_pmcid_and_pmid_for_doi(doi)
    assert id_dict == output


@pytest.mark.parametrize(
    ("doi", "id_dict"),
    [
        ("10.1098/rsif.2017.0387", {"PMCID": "PMC5938574", "PMID": "29618526"}),
        ("10.7554/ELIFE.32822", {"PMCID": "PMC5832410", "PMID": "29424689"}),
        (
            "10.1161/CIRCGENETICS.115.001181",
            {"PMID": "27094199"},
        ),  # only in PubMed, not in PMC
        ("10.7717/peerj.000", {}),  # Non-existent DOI
    ],
)
def test_get_pubmed_ids_for_doi(doi, id_dict):
    output = get_pubmed_ids_for_doi(doi)
    assert id_dict == output



import copy

import pytest

from manubot.cite.csl_item import CSL_Item, assert_csl_item_type


class Test_CSL_Item:
    def test_constuctor_empty(self):
        assert CSL_Item() == {}

    def test_constuctor_by_dict(self):
        d = {"title": "My book"}
        assert CSL_Item(d) == d

    def test_constuctor_by_keyword(self):
        assert CSL_Item(type="journal-article") == {"type": "journal-article"}

    def test_constuctor_by_dict_keyword_combination(self):
        assert CSL_Item({"title": "My journal article"}, type="journal-article") == {
            "title": "My journal article",
            "type": "journal-article",
        }

    def test_recursive_constructor(self):
        assert CSL_Item(CSL_Item()) == {}
        assert CSL_Item(CSL_Item(abc=1)) == {"abc": 1}

    def test_constructor_leaves_no_inplace_effects(self):
        dict1 = {"a": 1}
        ci = CSL_Item(dict1, b=2)
        assert ci == {"a": 1, "b": 2}
        assert dict1 == {"a": 1}

    def test_correct_invalid_type(self):
        assert CSL_Item(type="journal-article").correct_invalid_type() == {
            "type": "article-journal"
        }

    def test_set_default_type(self):
        assert CSL_Item().set_default_type() == {"type": "entry"}

    def test_no_change_of_type(self):
        assert CSL_Item(type="book").correct_invalid_type() == {"type": "book"}
        assert CSL_Item(type="book").set_default_type() == {"type": "book"}

    def test_clean(self):
        csl_item = CSL_Item(type="chapter", id="abc")
        csl_item.clean(prune=True)
        assert csl_item == {"type": "chapter", "id": "abc"}

    def test_clean_set_id(self):
        csl_item = CSL_Item(type="chapter")
        csl_item.set_id("abc")
        csl_item.clean(prune=True)
        assert csl_item == {"type": "chapter", "id": "abc"}


def test_assert_csl_item_type_passes():
    assert_csl_item_type(CSL_Item())


def test_assert_csl_item_type_raises_error_on_dict():
    with pytest.raises(TypeError):
        assert_csl_item_type({})


@pytest.mark.parametrize(
    ["csl_item", "standard_citation"],
    [
        (
            {"id": "my-id", "standard_citation": "doi:10.7554/elife.32822"},
            "doi:10.7554/elife.32822",
        ),
        ({"id": "doi:10.7554/elife.32822"}, "doi:10.7554/elife.32822"),
        ({"id": "doi:10.7554/ELIFE.32822"}, "doi:10.7554/elife.32822"),
        ({"id": "my-id"}, "raw:my-id"),
    ],
    ids=[
        "from_standard_citation",
        "from_doi_id",
        "from_doi_id_standardize",
        "from_raw_id",
    ],
)
def test_csl_item_standardize_id(csl_item, standard_citation):
    csl_item = CSL_Item(csl_item)
    output = csl_item.standardize_id()
    assert output is csl_item
    assert output["id"] == standard_citation


def test_csl_item_standardize_id_repeated():
    csl_item = CSL_Item(id="pmid:1", type="article-journal")
    csl_item_1 = copy.deepcopy(csl_item.standardize_id())
    assert "standard_citation" not in "csl_item"
    csl_item_2 = copy.deepcopy(csl_item.standardize_id())
    assert csl_item_1 == csl_item_2


def test_csl_item_standardize_id_note():
    """
    Test extracting standard_id from a note and setting additional
    note fields.
    """
    csl_item = CSL_Item(
        {
            "id": "original-id",
            "type": "article-journal",
            "note": "standard_id: doi:10.1371/journal.PPAT.1006256",
        }
    )
    csl_item.standardize_id()
    assert csl_item["id"] == "doi:10.1371/journal.ppat.1006256"
    note_dict = csl_item.note_dict
    assert note_dict["original_id"] == "original-id"
    assert note_dict["original_standard_id"] == "doi:10.1371/journal.PPAT.1006256"


@pytest.mark.parametrize(
    ["input_note", "text", "dictionary", "expected_note"],
    [
        ("", "", {}, ""),
        (None, "", {}, ""),
        ("preexisting note", "", {}, "preexisting note"),
        (
            "preexisting note",
            "",
            {"key": "the value"},
            "preexisting note\nkey: the value",
        ),
        ("", "", {"KEYOKAY": "the value"}, "KEYOKAY: the value"),
        ("preexisting note", "", {"KEY-NOT-OKAY": "the value"}, "preexisting note"),
        (
            "",
            "",
            {"standard_citation": "doi:10.7554/elife.32822"},
            "standard_citation: doi:10.7554/elife.32822",
        ),
        (
            "This CSL Item was produced using Manubot.",
            "",
            {"standard_citation": "doi:10.7554/elife.32822"},
            "This CSL Item was produced using Manubot.\nstandard_citation: doi:10.7554/elife.32822",
        ),
    ],
)
def test_csl_item_note_append(input_note, text, dictionary, expected_note):
    csl_item = CSL_Item({"id": "test_csl_item", "type": "entry", "note": input_note})
    csl_item.note_append_text(text)
    csl_item.note_append_dict(dictionary)
    assert csl_item.note == expected_note


@pytest.mark.parametrize(
    ["note", "dictionary"],
    [
        (
            "This is a note\nkey_one: value\nKEYTWO: value 2 ",
            {"key_one": "value", "KEYTWO": "value 2"},
        ),
        ("BAD_KEY: good value\ngood-key: good value", {"good-key": "good value"}),
        (
            "This is a note {:key_one: value} {:KEYTWO: value 2 } ",
            {"key_one": "value", "KEYTWO": "value 2"},
        ),
        ("{:BAD_KEY: good value}\n{:good-key: good value}", {"good-key": "good value"}),
        (
            "Mixed line-entry and braced-entry syntax\nGOODKEY: good value\n{:good-key: good value}",
            {"GOODKEY": "good value", "good-key": "good value"},
        ),
        ("Note without any key-value pairs", {}),
        (
            "Other text\nstandard_citation: doi:10/ckcj\nMore other text",
            {"standard_citation": "doi:10/ckcj"},
        ),
    ],
)
def test_csl_item_note_dict(note, dictionary):
    csl_item = CSL_Item(note=note)
    assert csl_item.note_dict == dictionary



import manubot.util

import pytest


def test_shlex_join():
    import pathlib

    args = ["command", "positional arg", "path_arg", pathlib.Path("path")]
    output = manubot.util.shlex_join(args)
    assert output == "command 'positional arg' path_arg path"


raw_repo_url = (
    "https://github.com/manubot/manubot/raw/ebac7abd754015a5ec24a6fff39c35a72d4dffb0/"
)
raw_manuscript_url = f"{raw_repo_url}manubot/process/tests/manuscripts/example/"


def test_read_serialized_data_url_yaml():
    url = raw_manuscript_url + "content/metadata.yaml"
    obj = manubot.util.read_serialized_data(url)
    assert obj["title"] == "Example manuscript for testing"
    obj = manubot.util.read_serialized_dict(url)
    assert obj["title"] == "Example manuscript for testing"


def test_read_serialized_data_url_json():
    url = raw_manuscript_url + "content/manual-references.json"
    obj = manubot.util.read_serialized_data(url)
    assert obj[0]["container-title"] == "Engineuring"
    with pytest.raises(TypeError, match="Received 'list' instead"):
        manubot.util.read_serialized_dict(url)


def test_read_serialized_dict_url_toml():
    url = raw_repo_url + "pyproject.toml"
    obj = manubot.util.read_serialized_dict(url)
    assert "black" in obj["tool"]



import collections
import functools
import json
import logging
import xml.etree.ElementTree

import requests

from manubot.util import get_manubot_user_agent


def get_pmc_csl_item(pmcid):
    """
    Get the CSL Item for a PubMed Central record by its PMID, PMCID, or
    DOI, using the NCBI Citation Exporter API.

    https://api.ncbi.nlm.nih.gov/lit/ctxp
    https://github.com/manubot/manubot/issues/21
    https://twitter.com/dhimmel/status/1061787168820092929
    """
    assert pmcid.startswith("PMC")
    csl_item = _get_literature_citation_exporter_csl_item("pmc", pmcid[3:])
    if "URL" not in csl_item:
        csl_item[
            "URL"
        ] = f"https://www.ncbi.nlm.nih.gov/pmc/articles/{csl_item.get('PMCID', pmcid)}/"
    return csl_item


def _get_literature_citation_exporter_csl_item(database, identifier):
    """
    https://api.ncbi.nlm.nih.gov/lit/ctxp
    """
    if database not in {"pubmed", "pmc"}:
        logging.error(
            f"Error calling _get_literature_citation_exporter_csl_item.\n"
            f'database must be either "pubmed" or "pmc", not {database}'
        )
        assert False
    if not identifier:
        logging.error(
            f"Error calling _get_literature_citation_exporter_csl_item.\n"
            f"identifier cannot be blank"
        )
        assert False
    params = {"format": "csl", "id": identifier}
    headers = {"User-Agent": get_manubot_user_agent()}
    url = f"https://api.ncbi.nlm.nih.gov/lit/ctxp/v1/{database}/"
    response = requests.get(url, params, headers=headers)
    try:
        csl_item = response.json()
    except Exception as error:
        logging.error(
            f"Error fetching {database} metadata for {identifier}.\n"
            f"Invalid JSON response from {response.url}:\n{response.text}"
        )
        raise error
    assert isinstance(csl_item, dict)
    if csl_item.get("status", "okay") == "error":
        logging.error(
            f"Error fetching {database} metadata for {identifier}.\n"
            f"Literature Citation Exporter returned JSON indicating an error for {response.url}\n"
            f"{json.dumps(csl_item, indent=2)}"
        )
        assert False
    return csl_item


def get_pubmed_csl_item(pmid):
    """
    Query NCBI E-Utilities to create CSL Items for PubMed IDs.

    https://github.com/manubot/manubot/issues/21
    https://github.com/ncbi/citation-exporter/issues/3#issuecomment-355313143
    """
    pmid = str(pmid)
    params = {"db": "pubmed", "id": pmid, "rettype": "full"}
    headers = {"User-Agent": get_manubot_user_agent()}
    url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi"
    with _get_eutils_rate_limiter():
        response = requests.get(url, params, headers=headers)
    try:
        element_tree = xml.etree.ElementTree.fromstring(response.text)
        (element_tree,) = list(element_tree)
    except Exception as error:
        logging.error(
            f"Error fetching PubMed metadata for {pmid}.\n"
            f"Invalid XML response from {response.url}:\n{response.text}"
        )
        raise error
    try:
        csl_item = csl_item_from_pubmed_article(element_tree)
    except Exception as error:
        msg = f"Error parsing the following PubMed metadata for PMID {pmid}:\n{response.text}"
        logging.error(msg)
        raise error
    return csl_item


def csl_item_from_pubmed_article(article):
    """
    article is a PubmedArticle xml element tree

    https://github.com/citation-style-language/schema/blob/master/csl-data.json
    """
    csl_item = collections.OrderedDict()

    if not article.find("MedlineCitation/Article"):
        raise NotImplementedError("Unsupported PubMed record: no <Article> element")

    title = article.findtext("MedlineCitation/Article/ArticleTitle")
    if title:
        csl_item["title"] = title

    volume = article.findtext("MedlineCitation/Article/Journal/JournalIssue/Volume")
    if volume:
        csl_item["volume"] = volume

    issue = article.findtext("MedlineCitation/Article/Journal/JournalIssue/Issue")
    if issue:
        csl_item["issue"] = issue

    page = article.findtext("MedlineCitation/Article/Pagination/MedlinePgn")
    if page:
        csl_item["page"] = page

    journal = article.findtext("MedlineCitation/Article/Journal/Title")
    if journal:
        csl_item["container-title"] = journal

    journal_short = article.findtext("MedlineCitation/Article/Journal/ISOAbbreviation")
    if journal_short:
        csl_item["container-title-short"] = journal_short

    issn = article.findtext("MedlineCitation/Article/Journal/ISSN")
    if issn:
        csl_item["ISSN"] = issn

    date_parts = extract_publication_date_parts(article)
    if date_parts:
        csl_item["issued"] = {"date-parts": [date_parts]}

    authors_csl = list()
    authors = article.findall("MedlineCitation/Article/AuthorList/Author")
    for author in authors:
        author_csl = collections.OrderedDict()
        given = author.findtext("ForeName")
        if given:
            author_csl["given"] = given
        family = author.findtext("LastName")
        if family:
            author_csl["family"] = family
        authors_csl.append(author_csl)
    if authors_csl:
        csl_item["author"] = authors_csl

    for id_type, key in ("pubmed", "PMID"), ("pmc", "PMCID"), ("doi", "DOI"):
        xpath = f"PubmedData/ArticleIdList/ArticleId[@IdType='{id_type}']"
        value = article.findtext(xpath)
        if value:
            csl_item[key] = value.lower() if key == "DOI" else value

    abstract = article.findtext("MedlineCitation/Article/Abstract/AbstractText")
    if abstract:
        csl_item["abstract"] = abstract

    csl_item["URL"] = f"https://www.ncbi.nlm.nih.gov/pubmed/{csl_item['PMID']}"
    csl_item["type"] = "article-journal"

    return csl_item


month_abbrev_to_int = {
    "Jan": 1,
    "Feb": 2,
    "Mar": 3,
    "Apr": 4,
    "May": 5,
    "Jun": 6,
    "Jul": 7,
    "Aug": 8,
    "Sep": 9,
    "Oct": 10,
    "Nov": 11,
    "Dec": 12,
}


def extract_publication_date_parts(article):
    """
    Extract date published from a PubmedArticle xml element tree.
    """
    date_parts = []

    # Electronic articles
    date = article.find("MedlineCitation/Article/ArticleDate")
    if date:
        for part in "Year", "Month", "Day":
            part = date.findtext(part)
            if not part:
                break
            date_parts.append(int(part))
        return date_parts

    # Print articles
    date = article.find("MedlineCitation/Article/Journal/JournalIssue/PubDate")
    year = date.findtext("Year")
    if year:
        date_parts.append(int(year))
    month = date.findtext("Month")
    if month:
        try:
            date_parts.append(month_abbrev_to_int[month])
        except KeyError:
            date_parts.append(int(month))
    day = date.findtext("Day")
    if day:
        date_parts.append(int(day))
    return date_parts


def get_pmcid_and_pmid_for_doi(doi):
    """
    Query PMC's ID Converter API to retrieve the PMCID and PMID for a DOI.
    Does not work for DOIs that are in Pubmed but not PubMed Central.
    https://www.ncbi.nlm.nih.gov/pmc/tools/id-converter-api/
    """
    assert isinstance(doi, str)
    assert doi.startswith("10.")
    params = {"ids": doi, "tool": "manubot"}
    url = "https://www.ncbi.nlm.nih.gov/pmc/utils/idconv/v1.0/"
    response = requests.get(url, params)
    if not response.ok:
        logging.warning(f"Status code {response.status_code} querying {response.url}\n")
        return {}
    try:
        element_tree = xml.etree.ElementTree.fromstring(response.text)
    except Exception:
        logging.warning(
            f"Error fetching PMC ID conversion for {doi}.\n"
            f"Response from {response.url}:\n{response.text}"
        )
        return {}
    records = element_tree.findall("record")
    if len(records) != 1:
        logging.warning(
            f"Expected PubMed Central ID converter to return a single XML record for {doi}.\n"
            f"Response from {response.url}:\n{response.text}"
        )
        return {}
    (record,) = records
    if record.findtext("status", default="okay") == "error":
        return {}
    id_dict = {}
    for id_type in "pmcid", "pmid":
        id_ = record.get(id_type)
        if id_:
            id_dict[id_type.upper()] = id_
    return id_dict


def get_pmid_for_doi(doi):
    """
    Query NCBI's E-utilities to retrieve the PMID for a DOI.
    """
    assert isinstance(doi, str)
    assert doi.startswith("10.")
    params = {"db": "pubmed", "term": f"{doi}[DOI]"}
    headers = {"User-Agent": get_manubot_user_agent()}
    url = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils/esearch.fcgi"
    with _get_eutils_rate_limiter():
        response = requests.get(url, params, headers=headers)
    if not response.ok:
        logging.warning(f"Status code {response.status_code} querying {response.url}\n")
        return None
    try:
        element_tree = xml.etree.ElementTree.fromstring(response.text)
    except Exception:
        logging.warning(
            f"Error in ESearch XML for DOI: {doi}.\n"
            f"Response from {response.url}:\n{response.text}"
        )
        return None
    id_elems = element_tree.findall("IdList/Id")
    if len(id_elems) != 1:
        logging.debug(
            f"No PMIDs found for {doi}.\n"
            f"Response from {response.url}:\n{response.text}"
        )
        return None
    (id_elem,) = id_elems
    return id_elem.text


def get_pubmed_ids_for_doi(doi):
    """
    Return a dictionary with PMCID and PMID, if they exist, for the specified
    DOI. See https://github.com/manubot/manubot/issues/45.
    """
    pubmed_ids = get_pmcid_and_pmid_for_doi(doi)
    if not pubmed_ids:
        pmid = get_pmid_for_doi(doi)
        if pmid:
            pubmed_ids["PMID"] = pmid
    return pubmed_ids


@functools.lru_cache()
def _get_eutils_rate_limiter():
    """
    Rate limiter to cap NCBI E-utilities queries to 3 per second as per
    https://ncbiinsights.ncbi.nlm.nih.gov/2017/11/02/new-api-keys-for-the-e-utilities/
    """
    from ratelimiter import RateLimiter

    return RateLimiter(max_calls=2, period=1)



"""
Manubot's command line interface
"""
import argparse
import logging
import pathlib
import sys
import warnings

import manubot
from manubot.util import import_function


def parse_arguments():
    """
    Read and process command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="Manubot: the manuscript bot for scholarly writing"
    )
    parser.add_argument(
        "--version", action="version", version=f"v{manubot.__version__}"
    )
    subparsers = parser.add_subparsers(
        title="subcommands", description="All operations are done through subcommands:"
    )
    # Require specifying a sub-command
    subparsers.required = True  # https://bugs.python.org/issue26510
    subparsers.dest = "subcommand"  # https://bugs.python.org/msg186387
    add_subparser_process(subparsers)
    add_subparser_cite(subparsers)
    add_subparser_webpage(subparsers)
    for subparser in subparsers.choices.values():
        subparser.add_argument(
            "--log-level",
            default="WARNING",
            choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
            help="Set the logging level for stderr logging",
        )
    args = parser.parse_args()
    return args


def add_subparser_process(subparsers):
    parser = subparsers.add_parser(
        name="process",
        help="process manuscript content",
        description="Process manuscript content to create outputs for Pandoc consumption. "
        "Performs bibliographic processing and templating.",
    )
    parser.add_argument(
        "--content-directory",
        type=pathlib.Path,
        required=True,
        help="Directory where manuscript content files are located.",
    )
    parser.add_argument(
        "--output-directory",
        type=pathlib.Path,
        required=True,
        help="Directory to output files generated by this script.",
    )
    parser.add_argument(
        "--template-variables-path",
        action="append",
        default=[],
        help="Path or URL of a file containing template variables for jinja2. "
        "Serialization format is inferred from the file extension, with support for JSON, YAML, and TOML. "
        "If the format cannot be detected, the parser assumes JSON. "
        "Specify this argument multiple times to read multiple files. "
        "Variables can be applied to a namespace (i.e. stored under a dictionary key) "
        "like `--template-variables-path=namespace=path_or_url`. "
        "Namespaces must match the regex `[a-zA-Z_][a-zA-Z0-9_]*`.",
    )
    parser.add_argument(
        "--cache-directory",
        type=pathlib.Path,
        help="Custom cache directory. " "If not specified, caches to output-directory.",
    )
    parser.add_argument("--clear-requests-cache", action="store_true")
    parser.set_defaults(function="manubot.process.process_command.cli_process")


def add_subparser_cite(subparsers):
    parser = subparsers.add_parser(
        name="cite",
        help="citation to CSL command line utility",
        description="Retrieve bibliographic metadata for one or more citation keys.",
    )
    parser.add_argument(
        "--render",
        action="store_true",
        help="Whether to render CSL Data into a formatted reference list using Pandoc. "
        "Pandoc version 2.0 or higher is required for complete support of available output formats.",
    )
    parser.add_argument(
        "--csl",
        default="https://github.com/greenelab/manubot-rootstock/raw/master/build/assets/style.csl",
        help="When --render, specify an XML CSL definition to style references (i.e. Pandoc's --csl option). "
        "Defaults to Manubot's style.",
    )
    parser.add_argument(
        "--format",
        choices=["plain", "markdown", "docx", "html", "jats"],
        help="When --render, format to use for output file. "
        "If not specified, attempt to infer this from filename extension. "
        "Otherwise, default to plain.",
    )
    parser.add_argument(
        "--output",
        type=pathlib.Path,
        help="Specify a file to write output, otherwise default to stdout.",
    )
    parser.add_argument(
        "--allow-invalid-csl-data",
        dest="prune_csl",
        action="store_false",
        help="Allow CSL Items that do not conform to the JSON Schema. Skips CSL pruning.",
    )
    parser.add_argument(
        "citekeys",
        nargs="+",
        help="One or more (space separated) citation keys to produce CSL for.",
    )
    parser.set_defaults(function="manubot.cite.cite_command.cli_cite")


def add_subparser_webpage(subparsers):
    parser = subparsers.add_parser(
        name="webpage",
        help="deploy Manubot outputs to a webpage directory tree",
        description="Update the webpage directory tree with Manubot output files. "
        "This command should be run from the root directory of a Manubot manuscript that follows the Rootstock layout, containing `output` and `webpage` directories. "
        "HTML and PDF outputs are copied to the webpage directory, which is structured as static source files for website hosting.",
    )
    parser.add_argument(
        "--checkout",
        nargs="?",
        const="gh-pages",
        default=None,
        help="branch to checkout /v directory contents from. "
        "For example, --checkout=upstream/gh-pages. "
        "--checkout is equivalent to --checkout=gh-pages. "
        "If --checkout is ommitted, no checkout is performed.",
    )
    parser.add_argument(
        "--version",
        help="Used to create webpage/v/{version} directory. "
        "Generally a commit hash, tag, or 'local'. "
        "When omitted, version defaults to the commit hash on CI builds and 'local' elsewhere.",
    )
    parser.add_argument(
        "--timestamp",
        action="store_true",
        help="timestamp versioned manuscripts in webpage/v using OpenTimestamps. "
        "Specify this flag to create timestamps for the current HTML and PDF outputs and upgrade any timestamps from past manuscript versions.",
    )
    cache_group = parser.add_mutually_exclusive_group()
    cache_group.add_argument(
        "--no-ots-cache", action="store_true", help="disable the timestamp cache."
    )
    cache_group.add_argument(
        "--ots-cache",
        default=pathlib.Path("ci/cache/ots"),
        type=pathlib.Path,
        help="location for the timestamp cache (default: ci/cache/ots).",
    )
    parser.set_defaults(function="manubot.webpage.webpage_command.cli_webpage")


def main():
    """
    Called as a console_scripts entry point in setup.py. This function defines
    the manubot command line script.
    """
    # Track if message gets logged with severity of error or greater
    # See https://stackoverflow.com/a/45446664/4651668
    import errorhandler

    error_handler = errorhandler.ErrorHandler()

    # Log DeprecationWarnings
    warnings.simplefilter("always", DeprecationWarning)
    logging.captureWarnings(True)

    # Log to stderr
    logger = logging.getLogger()
    stream_handler = logging.StreamHandler(stream=sys.stderr)
    stream_handler.setFormatter(
        logging.Formatter("## {levelname}\n{message}", style="{")
    )
    logger.addHandler(stream_handler)

    args = parse_arguments()
    logger.setLevel(getattr(logging, args.log_level))

    function = import_function(args.function)
    function(args)

    if error_handler.fired:
        logging.critical("Failure: exiting with code 1 due to logged errors")
        raise SystemExit(1)



import shutil

import pytest

from manubot.pandoc.tests.test_bibliography import bibliography_paths
from manubot.process.bibliography import load_manual_references


@pytest.mark.skipif(
    not shutil.which("pandoc-citeproc"),
    reason="pandoc-citeproc installation not found on system",
)
class Test_load_manual_references:
    """
    Tests loading multiple bibliography paths
    """

    def setup_method(self):
        self.citation_to_csl_item = load_manual_references(bibliography_paths)
        print(list(self.citation_to_csl_item))

    def test_csl_item_1(self):
        assert "doi:10.7554/elife.32822" in self.citation_to_csl_item
        csl_item_1 = self.citation_to_csl_item["doi:10.7554/elife.32822"]
        assert csl_item_1["title"].startswith("Sci-Hub")
        assert "CSL JSON Item was loaded by Manubot" in csl_item_1["note"]
        assert "manual_reference_filename: bibliography.json" in csl_item_1["note"]
        assert "standard_id: doi:10.7554/elife.32822" in csl_item_1["note"]

    def test_csl_item_2(self):
        # raw id corresponding to bibliography.bib
        assert "raw:noauthor_techblog:_nodate" in self.citation_to_csl_item
        csl_item_2 = self.citation_to_csl_item["raw:noauthor_techblog:_nodate"]
        assert csl_item_2["title"].startswith("TechBlog")
        assert "manual_reference_filename: bibliography.bib" in csl_item_2["note"]
        assert "original_id: noauthor_techblog:_nodate" in csl_item_2["note"]

    def test_csl_item_3(self):
        # id inferred by pandoc-citeproc during bib2json conversion of .nbib file
        assert "raw:Beaulieu-Jones2017" in self.citation_to_csl_item
        csl_item_3 = self.citation_to_csl_item["raw:Beaulieu-Jones2017"]
        assert csl_item_3["author"][0]["family"] == "Beaulieu-Jones"
        assert "manual_reference_filename: bibliography.nbib" in csl_item_3["note"]
        assert "original_id: Beaulieu-Jones2017" in csl_item_3["note"]



import logging
import pathlib
import shutil
import subprocess

from manubot.util import shlex_join


def cli_webpage(args):
    """
    Execute manubot webpage commands.
    args should be an argparse.Namespace object created by parser.parse_args.
    """
    configure_args(args)
    logging.debug(f"Running `manubot webpage` with the following args:\n{args}")
    if args.timestamp:
        ots_upgrade(args)
    create_version(args)


def configure_args(args):
    """
    Perform additional processing of arguments that is not handled by argparse.
    Derive additional variables and add them to args.
    For example, add directories to args and create them if neccessary.
    Note that versions_directory is the parent of version_directory.
    """
    args_dict = vars(args)

    # If --timestamp specified, check that opentimestamps-client is installed
    if args.timestamp:
        ots_executable_path = shutil.which("ots")
        if not ots_executable_path:
            logging.error(
                "manubot webpage --timestamp was specified but opentimestamps-client not found on system. "
                "Setting --timestamp=False. "
                "Fix this by installing https://pypi.org/project/opentimestamps-client/"
            )
            args_dict["timestamp"] = False

    # Directory where Manubot outputs reside
    args_dict["output_directory"] = pathlib.Path("output")

    # Set webpage directory
    args_dict["webpage_directory"] = pathlib.Path("webpage")
    args.webpage_directory.mkdir(exist_ok=True)

    # Create webpage/v directory (if it doesn't already exist)
    args_dict["versions_directory"] = args.webpage_directory.joinpath("v")
    args.versions_directory.mkdir(exist_ok=True)

    # Checkout existing version directories
    checkout_existing_versions(args)

    # Apply --version argument defaults
    if args.version is None:
        from manubot.process.ci import get_continuous_integration_parameters

        ci_params = get_continuous_integration_parameters()
        if ci_params:
            args_dict["version"] = ci_params.get("commit", "local")
        else:
            args_dict["version"] = "local"

    # Create empty webpage/v/version directory
    version_directory = args.versions_directory.joinpath(args.version)
    if version_directory.is_dir():
        logging.warning(
            f"{version_directory} exists: replacing it with an empty directory"
        )
        shutil.rmtree(version_directory)
    version_directory.mkdir()
    args_dict["version_directory"] = version_directory

    # Symlink webpage/v/latest to point to webpage/v/commit
    latest_directory = args.versions_directory.joinpath("latest")
    if latest_directory.is_symlink() or latest_directory.is_file():
        latest_directory.unlink()
    elif latest_directory.is_dir():
        shutil.rmtree(latest_directory)
    latest_directory.symlink_to(args.version, target_is_directory=True)
    args_dict["latest_directory"] = latest_directory

    # Create freeze directory
    freeze_directory = args.versions_directory.joinpath("freeze")
    freeze_directory.mkdir(exist_ok=True)
    args_dict["freeze_directory"] = freeze_directory

    return args


def checkout_existing_versions(args):
    """
    Must populate webpage/v from the gh-pages branch to get history
    References:
    http://clubmate.fi/git-checkout-file-or-directories-from-another-branch/
    https://stackoverflow.com/a/2668947/4651668
    https://stackoverflow.com/a/16493707/4651668
    Command modeled after:
    git --work-tree=webpage checkout upstream/gh-pages -- v
    """
    if not args.checkout:
        return
    command = [
        "git",
        f"--work-tree={args.webpage_directory}",
        "checkout",
        args.checkout,
        "--",
        "v",
    ]
    logging.info(
        f"Attempting checkout with the following command:\n{shlex_join(command)}"
    )
    process = subprocess.run(command, stderr=subprocess.PIPE)
    if process.returncode == 0:
        # Addresses an odd behavior where git checkout stages v/* files that don't actually exist
        subprocess.run(["git", "add", "v"])
    else:
        stderr = process.stderr.decode()
        message = (
            f"Checkout returned a nonzero exit status. See stderr:\n{stderr.rstrip()}"
        )
        if "pathspec" in stderr:
            message += (
                "\nManubot note: if there are no preexisting webpage versions (like for a newly created manuscript), "
                "the pathspec error above is expected and can be safely ignored."
            )  # see https://github.com/manubot/rootstock/issues/183
        logging.warning(message)


def create_version(args):
    """
    Populate the version directory for a new version.
    """
    # Copy content/images to webpage/v/commit/images
    images_src = pathlib.Path("content/images")
    if images_src.exists():
        shutil.copytree(src=images_src, dst=args.version_directory.joinpath("images"))

    # Copy output files to to webpage/v/version/
    renamer = {"manuscript.html": "index.html", "manuscript.pdf": "manuscript.pdf"}
    for src, dst in renamer.items():
        src_path = args.output_directory.joinpath(src)
        if not src_path.exists():
            continue
        dst_path = args.version_directory.joinpath(dst)
        shutil.copy2(src=src_path, dst=dst_path)
        if args.timestamp:
            ots_stamp(dst_path)

    # Create v/freeze to redirect to v/commit
    path = pathlib.Path(__file__).with_name("redirect-template.html")
    redirect_html = path.read_text()
    redirect_html = redirect_html.format(url=f"../{args.version}/")
    args.freeze_directory.joinpath("index.html").write_text(redirect_html)


def get_versions(args):
    """
    Extract versions from the webpage/v directory, which should each contain
    a manuscript.
    """
    versions = {x.name for x in args.versions_directory.iterdir() if x.is_dir()}
    versions -= {"freeze", "latest"}
    versions = sorted(versions)
    return versions


def ots_upgrade(args):
    """
    Upgrade OpenTimestamps .ots files in versioned commit directory trees.
    Upgrades each .ots file with a separate ots upgrade subprocess call due to
    https://github.com/opentimestamps/opentimestamps-client/issues/71
    """
    ots_paths = list()
    for version in get_versions(args):
        ots_paths.extend(args.versions_directory.joinpath(version).glob("**/*.ots"))
    ots_paths.sort()
    for ots_path in ots_paths:
        process_args = ["ots"]
        if args.no_ots_cache:
            process_args.append("--no-cache")
        else:
            process_args.extend(["--cache", str(args.ots_cache)])
        process_args.extend(["upgrade", str(ots_path)])
        process = subprocess.run(
            process_args, stderr=subprocess.PIPE, universal_newlines=True
        )
        message = f">>> {shlex_join(process.args)}\n{process.stderr}"
        if process.returncode != 0:
            logging.warning(
                f"OpenTimestamp upgrade failed with exit code {process.returncode}.\n{message}"
            )
        elif not process.stderr.strip() == "Success! Timestamp complete":
            logging.info(message)
        backup_path = ots_path.with_suffix(".ots.bak")
        if backup_path.exists():
            if process.returncode == 0:
                backup_path.unlink()
            else:
                # Restore original timestamp if failure
                backup_path.rename(ots_path)


def ots_stamp(path):
    """
    Timestamp a file using OpenTimestamps.
    This function calls `ots stamp path`.
    If `path` does not exist, this function does nothing.
    """
    process_args = ["ots", "stamp", str(path)]
    process = subprocess.run(
        process_args, stderr=subprocess.PIPE, universal_newlines=True
    )
    if process.returncode != 0:
        logging.warning(
            f"OpenTimestamp command returned nonzero code ({process.returncode}).\n"
            f">>> {shlex_join(process.args)}\n"
            f"{process.stderr}"
        )



def test_imports():
    import manubot.cite
    import manubot.cite.arxiv
    import manubot.cite.doi
    import manubot.cite.pubmed
    import manubot.cite.url
    import manubot.process.manuscript


def assert_instance_type():
    from manubot.cite.citekey import citeproc_retrievers

    assert isinstance(citeproc_retrievers, dict)



from manubot.cite.url import get_url_csl_item_zotero


def test_get_url_csl_item_zotero_nyt():
    """
    This command creates two translation-server queries. The first query is
    equivalent to:
    ```
    curl --verbose \
      --header "Content-Type: text/plain" \
      --data 'https://nyti.ms/1NuB0WJ' \
      'https://translate.manubot.org/web'
    ```
    """
    url = "https://nyti.ms/1NuB0WJ"
    csl_item = get_url_csl_item_zotero(url)
    assert (
        csl_item["title"] == "Unraveling the Ties of Altitude, Oxygen and Lung Cancer"
    )
    assert csl_item["author"][0]["family"] == "Johnson"


def test_get_url_csl_item_zotero_manubot():
    """
    This command creates two translation-server queries. The first query is
    equivalent to:
    ```
    curl --verbose \
      --header "Content-Type: text/plain" \
      --data 'https://greenelab.github.io/meta-review/v/0770300e1d5490a1ae8ff3a85ddca2cdc4ae0613/' \
      'https://translate.manubot.org/web'
    ```
    """
    url = "https://greenelab.github.io/meta-review/v/0770300e1d5490a1ae8ff3a85ddca2cdc4ae0613/"
    csl_item = get_url_csl_item_zotero(url)
    assert csl_item["title"] == "Open collaborative writing with Manubot"
    assert csl_item["author"][1]["family"] == "Slochower"
    # Zotero CSL exporter returns mixed string/int date-parts
    # https://github.com/zotero/zotero/issues/1603
    assert [int(x) for x in csl_item["issued"]["date-parts"][0]] == [2018, 12, 18]


def test_get_url_csl_item_zotero_github():
    """
    This command creates two translation-server queries. The first query is
    equivalent to:
    ```
    curl --verbose \
      --header "Content-Type: text/plain" \
      --data 'https://github.com/pandas-dev/pandas/tree/d5e5bf761092c59eeb9b8750f05f2bc29fb45927' \
      'https://translate.manubot.org/web'
    ```

    Note: this test may have temporary failures, due to performance of
          translation-server. It seems that sometimes translation-server
          returns a different title for the same URL. A real mystery.

    See also:
        https://github.com/manubot/manubot/pull/139#discussion_r328703233

    Proposed action:
        Probably should inquire upstream or change the test.
    """
    url = "https://github.com/pandas-dev/pandas/tree/d5e5bf761092c59eeb9b8750f05f2bc29fb45927"
    csl_item = get_url_csl_item_zotero(url)
    # FIXME: arbitraryly, csl_item['abstract'], and not csl_item['title'] contains the title.
    assert csl_item["title"].startswith("Flexible and powerful data analysis")
    assert csl_item["source"] == "GitHub"



import collections
import json
import logging
import re
import textwrap
import warnings
from typing import List, Optional

import jinja2
import pandas
import requests
import requests_cache
import yaml

from manubot.util import read_serialized_data, read_serialized_dict
from manubot.process.bibliography import load_manual_references
from manubot.process.ci import get_continuous_integration_parameters
from manubot.process.metadata import (
    get_header_includes,
    get_thumbnail_url,
    get_manuscript_urls,
)
from manubot.process.manuscript import (
    datetime_now,
    get_citekeys,
    get_manuscript_stats,
    get_text,
    update_manuscript_citekeys,
)
from manubot.cite.citekey import (
    citekey_to_csl_item,
    shorten_citekey,
    is_valid_citekey,
    standardize_citekey,
)


def check_collisions(citekeys_df):
    """
    Check for short_citekey hash collisions
    """
    collision_df = citekeys_df[["standard_citekey", "short_citekey"]].drop_duplicates()
    collision_df = collision_df[collision_df.short_citekey.duplicated(keep=False)]
    if not collision_df.empty:
        logging.error(f"OMF! Hash collision. Congratulations.\n{collision_df}")
    return collision_df


def check_multiple_citation_strings(citekeys_df):
    """
    Identify different citation strings referring the the same reference.
    """
    message = textwrap.dedent(
        f"""\
    {len(citekeys_df)} unique citations strings extracted from text
    {citekeys_df.standard_citekey.nunique()} unique standard citations\
    """
    )
    logging.info(message)
    multi_df = citekeys_df[citekeys_df.standard_citekey.duplicated(keep=False)]
    if not multi_df.empty:
        table = multi_df.to_string(
            index=False, columns=["standard_citekey", "manuscript_citekey"]
        )
        logging.warning(f"Multiple citekeys detected for the same reference:\n{table}")
    return multi_df


def read_variable_files(paths: List[str], variables: Optional[dict] = None) -> dict:
    """
    Read multiple serialized data files into a user_variables dictionary.
    Provide `paths` (a list of URLs or local file paths).
    Paths can optionally have a namespace prepended.
    For example:

    ```python
    paths = [
        'https://git.io/vbkqm',  # update the dictionary's top-level
        'namespace_1=https://git.io/vbkqm',  # store under 'namespace_1' key
        'namespace_2=some_local_path.json',  # store under 'namespace_2' key
    ]
    ```

    If a namespace is not provided, the JSON must contain a dictionary as its
    top level. Namespaces should consist only of ASCII alphanumeric characters
    (includes underscores, first character cannot be numeric).

    Pass a dictionary to `variables` to update an existing dictionary rather
    than create a new dictionary.
    """
    if variables is None:
        variables = {}
    for path in paths:
        logging.info(f"Reading user-provided templating variables at {path!r}")
        # Match only namespaces that are valid jinja2 variable names
        # http://jinja.pocoo.org/docs/2.10/api/#identifier-naming
        match = re.match(r"([a-zA-Z_][a-zA-Z0-9_]*)=(.+)", path)
        if match:
            namespace, path = match.groups()
            logging.info(
                f"Using the {namespace!r} namespace for template variables from {path!r}"
            )
        try:
            if match:
                obj = {namespace: read_serialized_data(path)}
            else:
                obj = read_serialized_dict(path)
        except Exception:
            logging.exception(f"Error reading template variables from {path!r}")
            continue
        assert isinstance(obj, dict)
        conflicts = variables.keys() & obj.keys()
        if conflicts:
            logging.warning(
                f"Template variables in {path!r} overwrite existing "
                "values for the following keys:\n" + "\n".join(conflicts)
            )
        variables.update(obj)
    logging.debug(
        f"Reading user-provided templating variables complete:\n"
        f"{json.dumps(variables, indent=2, ensure_ascii=False)}"
    )
    return variables


def add_author_affiliations(variables: dict) -> dict:
    """
    Edit variables to contain numbered author affiliations. Specifically,
    add a list of affiliation_numbers for each author and add a list of
    affiliations to the top-level of variables. If no authors have any
    affiliations, variables is left unmodified.
    """
    rows = list()
    for author in variables["authors"]:
        if "affiliations" not in author:
            continue
        if not isinstance(author["affiliations"], list):
            warnings.warn(
                f"Expected list for {author['name']}'s affiliations. "
                f"Assuming multiple affiliations are `; ` separated. "
                f"Please switch affiliations to a list.",
                category=DeprecationWarning,
            )
            author["affiliations"] = author["affiliations"].split("; ")
        for affiliation in author["affiliations"]:
            rows.append((author["name"], affiliation))
    if not rows:
        return variables
    affil_map_df = pandas.DataFrame(rows, columns=["name", "affiliation"])
    affiliation_df = affil_map_df[["affiliation"]].drop_duplicates()
    affiliation_df["affiliation_number"] = range(1, 1 + len(affiliation_df))
    affil_map_df = affil_map_df.merge(affiliation_df)
    name_to_numbers = {
        name: sorted(df.affiliation_number) for name, df in affil_map_df.groupby("name")
    }
    for author in variables["authors"]:
        author["affiliation_numbers"] = name_to_numbers.get(author["name"], [])
    variables["affiliations"] = affiliation_df.to_dict(orient="records")
    return variables


def load_variables(args) -> dict:
    """
    Read `metadata.yaml` and files specified by `--template-variables-path` to generate
    manuscript variables available for jinja2 templating.

    Returns a dictionary, refered to as `variables`, with the following keys:

    - `pandoc`: a dictionary for passing options to Pandoc via the `yaml_metadata_block`.
      Fields in `pandoc` are either generated by Manubot or hard-coded by the user if `metadata.yaml`
      includes a `pandoc` dictionary.
    - `manubot`: a dictionary for manubot-related information and metadata.
      Fields in `manubot` are either generated by Manubot or hard-coded by the user if `metadata.yaml`
      includes a `manubot` dictionary.
    - All fields from a manuscript's `metadata.yaml` that are not interpreted by Manubot are
      copied to `variables`. Interpreted fields include `pandoc`, `manubot`, `title`,
      `keywords`, `author_info`, `lang`, and `thumbnail`.
    - User-specified fields inserted according to the `--template-variables-path` option.
      User-specified variables take highest precedence and can overwrite values for existing
      keys like `pandoc` or `manubot` (dangerous).
    """
    # Generated manuscript variables
    variables = {"pandoc": {}, "manubot": {}}

    # Read metadata which contains pandoc_yaml_metadata
    # as well as author_info.
    if args.meta_yaml_path.is_file():
        metadata = read_serialized_dict(args.meta_yaml_path)
    else:
        metadata = {}
        logging.warning(
            f"missing {args.meta_yaml_path} file with yaml_metadata_block for pandoc"
        )

    # Interpreted keys that are intended for pandoc
    move_to_pandoc = "title", "keywords", "lang"
    for key in move_to_pandoc:
        if key in metadata:
            variables["pandoc"][key] = metadata.pop(key)

    # Add date to metadata
    now = datetime_now()
    logging.info(
        f"Using {now:%Z} timezone.\n"
        f"Dating manuscript with the current datetime: {now.isoformat()}"
    )
    variables["pandoc"]["date-meta"] = now.date().isoformat()
    variables["manubot"]["date"] = f"{now:%B} {now.day}, {now.year}"

    # Process authors metadata
    authors = metadata.pop("author_info", [])
    if authors is None:
        authors = []
    variables["pandoc"]["author-meta"] = [author["name"] for author in authors]
    variables["manubot"]["authors"] = authors
    add_author_affiliations(variables["manubot"])

    # Set repository version metadata for CI builds
    ci_params = get_continuous_integration_parameters()
    if ci_params:
        variables["manubot"]["ci_source"] = ci_params

    # Add manuscript URLs
    variables["manubot"].update(get_manuscript_urls(metadata.pop("html_url", None)))

    # Add thumbnail URL if present
    thumbnail_url = get_thumbnail_url(metadata.pop("thumbnail", None))
    if thumbnail_url:
        variables["manubot"]["thumbnail_url"] = thumbnail_url

    # Update variables with metadata.yaml pandoc/manubot dicts
    for key in "pandoc", "manubot":
        dict_ = metadata.pop(key, {})
        if not isinstance(dict_, dict):
            logging.warning(
                f"load_variables expected metadata.yaml field {key!r} to be a dict."
                f"Received a {dict_.__class__.__name__!r} instead."
            )
            continue
        variables[key].update(dict_)

    # Update variables with uninterpreted metadata.yaml fields
    variables.update(metadata)

    # Update variables with user-provided variables here
    variables = read_variable_files(args.template_variables_path, variables)

    # Add header-includes metadata with <meta> information for the HTML output's <head>
    variables["pandoc"]["header-includes"] = get_header_includes(variables)

    return variables


def get_citekeys_df(args, text):
    """
    Generate citekeys_df and save it to 'citations.tsv'.
    citekeys_df is a pandas.DataFrame with the following columns:
    - manuscript_citekey: citation keys extracted from the manuscript content files.
    - detagged_citekey: manuscript_citekey but with tag citekeys dereferenced
    - standard_citekey: detagged_citekey standardized
    - short_citekey: standard_citekey hashed to create a shortened citekey
    """
    citekeys_df = pandas.DataFrame({"manuscript_citekey": get_citekeys(text)})
    if args.citation_tags_path.is_file():
        tag_df = pandas.read_csv(args.citation_tags_path, sep="\t")
        na_rows_df = tag_df[tag_df.isnull().any(axis="columns")]
        if not na_rows_df.empty:
            logging.error(
                f"{args.citation_tags_path} contains rows with missing values:\n"
                f"{na_rows_df}\n"
                "This error can be caused by using spaces rather than tabs to delimit fields.\n"
                "Proceeding to reread TSV with delim_whitespace=True."
            )
            tag_df = pandas.read_csv(args.citation_tags_path, delim_whitespace=True)
        tag_df["manuscript_citekey"] = "tag:" + tag_df.tag
        tag_df = tag_df.rename(columns={"citation": "detagged_citekey"})
        for detagged_citekey in tag_df.detagged_citekey:
            is_valid_citekey(detagged_citekey, allow_raw=True)
        citekeys_df = citekeys_df.merge(
            tag_df[["manuscript_citekey", "detagged_citekey"]], how="left"
        )
    else:
        citekeys_df["detagged_citekey"] = None
        logging.info(
            f"missing {args.citation_tags_path} file: no citation tags (citekey aliases) set"
        )
    citekeys_df.detagged_citekey.fillna(
        citekeys_df.manuscript_citekey.astype(str), inplace=True
    )
    citekeys_df["standard_citekey"] = citekeys_df.detagged_citekey.map(
        standardize_citekey
    )
    citekeys_df["short_citekey"] = citekeys_df.standard_citekey.map(shorten_citekey)
    citekeys_df = citekeys_df.sort_values(["standard_citekey", "detagged_citekey"])
    citekeys_df.to_csv(args.citations_path, sep="\t", index=False)
    check_collisions(citekeys_df)
    check_multiple_citation_strings(citekeys_df)
    return citekeys_df


def generate_csl_items(args, citekeys_df):
    """
    General CSL (citeproc) items for standard_citekeys in citekeys_df.
    Writes references.json to disk and logs warnings for potential problems.
    """
    # Read manual references (overrides) in JSON CSL
    manual_refs = load_manual_references(args.manual_references_paths)

    requests  # require `import requests` in case this is essential for monkey patching by requests_cache.
    requests_cache.install_cache(args.requests_cache_path, include_get_headers=True)
    cache = requests_cache.get_cache()
    if args.clear_requests_cache:
        logging.info("Clearing requests-cache")
        requests_cache.clear()
    logging.info(
        f"requests-cache starting with {len(cache.responses)} cached responses"
    )

    csl_items = list()
    failures = list()
    for standard_citekey in citekeys_df.standard_citekey.unique():
        if standard_citekey in manual_refs:
            csl_items.append(manual_refs[standard_citekey])
            continue
        elif standard_citekey.startswith("raw:"):
            logging.error(
                f"CSL JSON Data with a standard_citekey of {standard_citekey!r} not found in manual-references.json. "
                "Metadata must be provided for raw citekeys."
            )
            failures.append(standard_citekey)
        try:
            csl_item = citekey_to_csl_item(standard_citekey)
            csl_items.append(csl_item)
        except Exception:
            logging.exception(f"Citeproc retrieval failure for {standard_citekey!r}")
            failures.append(standard_citekey)

    logging.info(
        f"requests-cache finished with {len(cache.responses)} cached responses"
    )
    requests_cache.uninstall_cache()

    if failures:
        message = "CSL JSON Data retrieval failed for the following standardized citation keys:\n{}".format(
            "\n".join(failures)
        )
        logging.error(message)

    # Write JSON CSL bibliography for Pandoc.
    with args.references_path.open("w", encoding="utf-8") as write_file:
        json.dump(csl_items, write_file, indent=2, ensure_ascii=False)
        write_file.write("\n")
    return csl_items


def template_with_jinja2(text, variables):
    """
    Template using jinja2 with the variables dictionary unpacked as keyword
    arguments.
    """
    jinja_environment = jinja2.Environment(
        loader=jinja2.BaseLoader(),
        undefined=jinja2.make_logging_undefined(logging.getLogger()),
        autoescape=False,
        comment_start_string="{##",
        comment_end_string="##}",
        extensions=["jinja2.ext.do", "jinja2.ext.loopcontrols"],
    )
    template = jinja_environment.from_string(text)
    return template.render(**variables)


def prepare_manuscript(args):
    """
    Compile manuscript, creating manuscript.md and references.json as inputs
    for pandoc.
    """
    text = get_text(args.content_directory)
    citekeys_df = get_citekeys_df(args, text)

    generate_csl_items(args, citekeys_df)

    citekey_mapping = collections.OrderedDict(
        zip(citekeys_df.manuscript_citekey, citekeys_df.short_citekey)
    )
    text = update_manuscript_citekeys(text, citekey_mapping)

    variables = load_variables(args)
    variables["manubot"]["manuscript_stats"] = get_manuscript_stats(text, citekeys_df)
    with args.variables_path.open("w", encoding="utf-8") as write_file:
        json.dump(variables, write_file, ensure_ascii=False, indent=2)
        write_file.write("\n")

    text = template_with_jinja2(text, variables)

    # Write manuscript for pandoc
    with args.manuscript_path.open("w", encoding="utf-8") as write_file:
        yaml.dump(
            variables["pandoc"],
            write_file,
            default_flow_style=False,
            explicit_start=True,
            explicit_end=True,
            width=float("inf"),
        )
        write_file.write("\n")
        write_file.write(text)



import collections
import json
import logging
import re


def get_isbn_csl_item(isbn):
    """
    Generate CSL JSON Data for an ISBN. Converts all ISBNs to 13-digit format.

    This function uses a list of CSL JSON Item metadata retrievers, specified
    by the module-level variable `isbn_retrievers`. The methods are attempted
    in order, with this function returning the metadata from the first
    non-failing method.
    """
    import isbnlib

    isbn = isbnlib.to_isbn13(isbn)
    for retriever in isbn_retrievers:
        try:
            return retriever(isbn)
        except Exception as error:
            logging.warning(
                f"Error in {retriever.__name__} for {isbn} "
                f"due to a {error.__class__.__name__}:\n{error}"
            )
            logging.info(error, exc_info=True)
    raise Exception(f"all get_isbn_csl_item methods failed for {isbn}")


def get_isbn_csl_item_zotero(isbn):
    """
    Generate CSL JSON Data for an ISBN using Zotero's translation-server.
    """
    from manubot.cite.zotero import get_csl_item

    return get_csl_item(f"isbn:{isbn}")


def get_isbn_csl_item_citoid(isbn):
    """
    Return CSL JSON Data for an ISBN using the Wikipedia Citoid API.
    https://en.wikipedia.org/api/rest_v1/#!/Citation/getCitation
    """
    import requests
    from manubot.util import get_manubot_user_agent

    headers = {"User-Agent": get_manubot_user_agent()}
    url = f"https://en.wikipedia.org/api/rest_v1/data/citation/mediawiki/{isbn}"
    response = requests.get(url, headers=headers)
    result = response.json()
    if isinstance(result, dict):
        if result["title"] == "Not found.":
            raise KeyError(f"Metadata for ISBN {isbn} not found at {url}")
        else:
            raise Exception(
                f"Unable to extract CSL from JSON metadata for ISBN {isbn}:\n"
                f"{json.dumps(result.text)}"
            )
    (mediawiki,) = result
    csl_item = collections.OrderedDict()
    csl_item["type"] = mediawiki.get("itemType", "book")
    if "title" in mediawiki:
        csl_item["title"] = mediawiki["title"]
    if "author" in mediawiki:
        csl_author = list()
        for last, first in mediawiki["author"]:
            csl_author.append({"given": first, "family": last})
        if csl_author:
            csl_item["author"] = csl_author
    if "date" in mediawiki:
        year_pattern = re.compile(r"[0-9]{4}")
        match = year_pattern.search(mediawiki["date"])
        if match:
            year = int(match.group())
            csl_item["issued"] = {"date-parts": [[year]]}
        else:
            logging.debug(
                f"get_isbn_csl_item_citoid: issue extracting date for ISBN {isbn}\n"
                f"metadata retrieved from {url}\n"
                f'unable to extract year from date field: {mediawiki["date"]}'
            )
    if "publisher" in mediawiki:
        csl_item["publisher"] = mediawiki["publisher"]
    if "place" in mediawiki:
        csl_item["publisher-place"] = mediawiki["place"]
    if "volume" in mediawiki:
        csl_item["volume"] = mediawiki["volume"]
    if "edition" in mediawiki:
        csl_item["edition"] = mediawiki["edition"]
    if "abstractNote" in mediawiki:
        csl_item["abstract"] = mediawiki["abstractNote"]
    csl_item["ISBN"] = isbn
    if "source" in mediawiki:
        csl_item["source"] = mediawiki["source"][0]
    if "url" in mediawiki:
        csl_item["URL"] = mediawiki["url"]
    return csl_item


def get_isbn_csl_item_isbnlib(isbn):
    """
    Generate CSL JSON Data for an ISBN using isbnlib.
    """
    import isbnlib

    metadata = isbnlib.meta(isbn, cache=None)
    csl_json = isbnlib.registry.bibformatters["csl"](metadata)
    csl_data = json.loads(csl_json)
    return csl_data


isbn_retrievers = [
    get_isbn_csl_item_zotero,
    get_isbn_csl_item_citoid,
    get_isbn_csl_item_isbnlib,
]



import pathlib

import pytest

from ..util import add_author_affiliations, read_variable_files

directory = pathlib.Path(__file__).parent.resolve()


def test_read_variable_files_empty():
    paths = []
    user_variables = read_variable_files(paths)
    assert isinstance(user_variables, dict) and not user_variables


def test_read_variable_files():
    """
    Test reading multiple JSON files, from both local paths and URLs.
    """
    local_path = "manuscripts/variables/content/template-variables.json"
    local_path = directory.joinpath(local_path)
    paths = [
        "https://git.io/vbkqm",
        "https://git.io/vbkqm",
        "namespace_1=https://git.io/vbkqm",
        "namespace_2=https://git.io/vbkqm",
        f"namespace_2={local_path}",
        f"namespace_3={local_path}",
    ]
    user_variables = read_variable_files(paths)
    assert "namespace_1" in user_variables
    assert "namespace_2" in user_variables
    assert "namespace_3" in user_variables
    assert user_variables["generated_by"] == "Manubot"
    assert "violet" in user_variables["namespace_1"]["rainbow"]
    assert "yellow" in user_variables["namespace_2"]["rainbow"]
    assert "orange" in user_variables["namespace_3"]["rainbow"]


def test_add_author_affiliations_empty():
    variables = {}
    variables["authors"] = [{"name": "Jane Roe"}, {"name": "John Doe"}]
    returned_variables = add_author_affiliations(variables)
    assert variables is returned_variables
    assert "affiliations" not in variables
    for author in variables["authors"]:
        assert "affiliation_numbers" not in author


def test_add_author_affiliations():
    variables = {}
    variables["authors"] = [
        # Deprecated affiliations format (as a string that's `; ` separated)
        {
            "name": "Jane Roe",
            "affiliations": "Department of Doe, University of Roe; Peppertea University",
        },
        # Prefered affiliations format as a list
        {
            "name": "John Doe",
            "affiliations": ["Unique University", "Peppertea University"],
        },
    ]
    with pytest.warns(DeprecationWarning):
        returned_variables = add_author_affiliations(variables)
    assert variables is returned_variables
    assert variables["affiliations"] == [
        {
            "affiliation": "Department of Doe, University of Roe",
            "affiliation_number": 1,
        },
        {"affiliation": "Peppertea University", "affiliation_number": 2},
        {"affiliation": "Unique University", "affiliation_number": 3},
    ]
    authors = variables["authors"]
    assert authors[0]["affiliations"] == [
        "Department of Doe, University of Roe",
        "Peppertea University",
    ]
    assert authors[0]["affiliation_numbers"] == [1, 2]
    assert authors[1]["affiliation_numbers"] == [2, 3]



import pathlib
import subprocess

import pytest

directory = pathlib.Path(__file__).parent.resolve()

# List of manuscripts for testing. All subdirectories of ./manuscripts
manuscripts = [
    path.name for path in directory.joinpath("manuscripts").iterdir() if path.is_dir()
]


@pytest.mark.parametrize("manuscript", manuscripts)
def test_example_manuscript(manuscript):
    """
    Test command line execution of manubot to build an example manuscript.
    """
    manuscript_dir = directory.joinpath("manuscripts", manuscript)
    args = [
        "manubot",
        "process",
        "--log-level",
        "INFO",
        "--content-directory",
        str(manuscript_dir.joinpath("content")),
        "--output-directory",
        str(manuscript_dir.joinpath("output")),
    ]
    if manuscript == "variables":
        args.extend(
            [
                "--template-variables-path",
                str(manuscript_dir.joinpath("content/template-variables.json")),
            ]
        )
    process = subprocess.run(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    print(process.args)
    print(process.stderr.decode())
    assert process.returncode == 0



import pytest

from manubot.cite.doi import (
    expand_short_doi,
    get_doi_csl_item,
    get_doi_csl_item_zotero,
    get_doi_csl_item_crosscite,
)


def test_expand_short_doi():
    doi = expand_short_doi("10/b6vnmd")
    assert doi == "10.1016/s0933-3657(96)00367-3"


def test_expand_short_doi_invalid():
    with pytest.raises(ValueError, match="Handle not found. Double check short_doi"):
        expand_short_doi("10/b6vnmdxxxxxx")


def test_expand_short_doi_not_short():
    with pytest.raises(ValueError, match="shortDOIs start with `10/`"):
        expand_short_doi("10.1016/S0933-3657(96)00367-3")


def test_get_doi_csl_item_crosscite():
    doi = "10.1101/142760"
    csl_item = get_doi_csl_item_crosscite(doi)
    assert isinstance(csl_item, dict)
    csl_item["publisher"] == "Cold Spring Harbor Laboratory"


def test_get_doi_csl_item_zotero():
    """
    As of 2019-10-25, DOI Content Negotiation (i.e. crosscite) encodes the consortium
    in author.name rather than author.literal. Zotero translation-server encodes
    the consortium in author.family, which is better than nothing since it's a valid
    CSL JSON field.
    https://github.com/manubot/manubot/issues/158
    """
    doi = "10.1038/ng.3834"
    csl_item = get_doi_csl_item_zotero(doi)
    assert isinstance(csl_item, dict)
    csl_item["author"][9]["family"] == "GTEx Consortium"


def test_get_doi_csl_item():
    """
    Test URL is set with shortDOI when calling get_doi_csl_item.
    """
    doi = "10.1101/142760"
    csl_item = get_doi_csl_item(doi)
    assert isinstance(csl_item, dict)
    assert csl_item["URL"] == "https://doi.org/gbpvh5"



import itertools
import pathlib
import re

import setuptools

directory = pathlib.Path(__file__).parent.resolve()

# version
init_path = directory.joinpath("manubot", "__init__.py")
text = init_path.read_text(encoding="utf-8-sig")
pattern = re.compile(r"^__version__ = ['\"]([^'\"]*)['\"]", re.MULTILINE)
version = pattern.search(text).group(1)

# long_description
readme_path = directory.joinpath("README.md")
long_description = readme_path.read_text(encoding="utf-8-sig")

# extra dependencies with an "all" option
extras_require = {
    "webpage": ["opentimestamps-client"],
    "test": ["black", "ghp-import", "portray", "pytest"],
}
extras_require["all"] = list(
    dict.fromkeys(itertools.chain.from_iterable(extras_require.values()))
)

setuptools.setup(
    # Package details
    name="manubot",
    version=version,
    url="https://github.com/manubot/manubot",
    project_urls={
        "Source": "https://github.com/manubot/manubot",
        "Documentation": "https://manubot.github.io/manubot",
        "Tracker": "https://github.com/manubot/manubot/issues",
        "Homepage": "https://manubot.org",
        "Publication": "https://greenelab.github.io/meta-review/",
    },
    description="Python utilities for Manubot: Manuscripts, open and automated",
    long_description_content_type="text/markdown",
    long_description=long_description,
    license="BSD 3-Clause",
    # Author details
    author="Daniel Himmelstein",
    author_email="daniel.himmelstein@gmail.com",
    # Package topics
    keywords="manuscript markdown publishing references citations",
    classifiers=[
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=setuptools.find_packages(exclude=["tests"]),
    # Specify python version
    python_requires=">=3.6",
    # Run-time dependencies
    install_requires=[
        "errorhandler",
        "isbnlib",
        "jinja2",
        "jsonref",
        "jsonschema",
        "packaging",
        "pandas",
        "pybase62",
        "pyyaml",
        "ratelimiter",
        "requests-cache",
        "requests",
        "toml",
    ],
    # Additional groups of dependencies
    extras_require=extras_require,
    # Create command line script
    entry_points={"console_scripts": ["manubot = manubot.command:main"]},
    # Include package data files
    package_data={
        "manubot": [
            "cite/*.lua",
            "process/header-includes-template.html",
            "webpage/redirect-template.html",
        ]
    },
)



import importlib
import json
import logging
import platform
import shlex
import sys

# Email address that forwards to Manubot maintainers
contact_email = "contact@manubot.org"


def import_function(name):
    """
    Import a function in a module specified by name. For example, if name were
    'manubot.cite.cite_command.cli_cite', the cli_cite function would be
    returned as an object. See https://stackoverflow.com/a/8790232/4651668.
    """
    module_name, function_name = name.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, function_name)


def get_manubot_user_agent():
    """
    Return a User-Agent string for web request headers to help services
    identify requests as coming from Manubot.
    """
    try:
        from manubot import __version__ as manubot_version
    except ImportError:
        manubot_version = ""
    return (
        f"manubot/{manubot_version} "
        f"({platform.system()}; Python/{sys.version_info.major}.{sys.version_info.minor}) "
        f"<{contact_email}>"
    )


def shlex_join(split_command):
    """
    Backport shlex.join for Python < 3.8.
    Also cast all args to str to increase versatility.
    https://github.com/python/cpython/pull/7605
    https://bugs.python.org/issue22454
    """
    return " ".join(shlex.quote(str(arg)) for arg in split_command)


"""Valid schemes for HTTP URL detection"""
_http_url_schemes = {"http", "https"}


def is_http_url(string: str) -> bool:
    """
    Return whether `string` is an HTTP(s) Uniform Resource Locator (URL).
    """
    from urllib.parse import urlparse

    parsed_url = urlparse(string)
    return parsed_url.scheme in _http_url_schemes


def read_serialized_data(path: str):
    """
    Read seralized data from a local file path or web-address.
    If file format extension is not detected in path, assumes JSON.
    If a URL does not contain the appropriate suffix, one workaround
    is to hack the fragment like https://example.org#/variables.toml
    """
    import os
    import pathlib
    import requests

    path_str = os.fspath(path)
    path_lib = pathlib.Path(path)
    supported_suffixes = {".json", ".yaml", ".yml", ".toml"}
    suffixes = set(path_lib.suffixes)
    if is_http_url(path_str):
        response = requests.get(path_str)
        if not suffixes & supported_suffixes:
            # if URL has no supported suffixes, evaluate suffixes of final redirect
            suffixes = set(pathlib.Path(response.url).suffixes)
        text = response.text
    else:
        text = path_lib.read_text(encoding="utf-8-sig")
    if {".yaml", ".yml"} & suffixes:
        import yaml

        return yaml.safe_load(text)
    if ".toml" in suffixes:
        import toml

        return toml.loads(text)
    if ".json" not in suffixes:
        logging.info(
            f"read_serialized_data cannot infer serialization format from the extension of {path_str!r}. "
            f"Supported extensions are {', '.join(supported_suffixes)}. "
            "Assuming JSON."
        )
    return json.loads(text)


def read_serialized_dict(path: str) -> dict:
    """
    Read serialized data, confirming that the top-level object is a dictionary.
    Delegates to `read_serialized_data`.
    """
    data = read_serialized_data(path)
    if isinstance(data, dict):
        return data
    raise TypeError(
        f"Expected data encoded by {path!r} to be a dictionary at the top-level. "
        f"Received {data.__class__.__name__!r} instead."
    )



"""Functions importable from manubot.cite submodule (submodule API):

  standardize_citekey()
  citekey_to_csl_item()

Helpers:

  inspect_citekey()
  is_valid_citekey() - also used in manubot.process
  shorten_citekey() - used solely in manubot.process
  infer_citekey_prefix()

"""
import functools
import logging
import re

from manubot.util import import_function

citeproc_retrievers = {
    "doi": "manubot.cite.doi.get_doi_csl_item",
    "pmid": "manubot.cite.pubmed.get_pubmed_csl_item",
    "pmcid": "manubot.cite.pubmed.get_pmc_csl_item",
    "arxiv": "manubot.cite.arxiv.get_arxiv_csl_item",
    "isbn": "manubot.cite.isbn.get_isbn_csl_item",
    "wikidata": "manubot.cite.wikidata.get_wikidata_csl_item",
    "url": "manubot.cite.url.get_url_csl_item",
}

"""
Regex to extract citation keys.
The leading '@' is omitted from the single match group.

Same rules as pandoc, except more permissive in the following ways:

1. the final character can be a slash because many URLs end in a slash.
2. underscores are allowed in internal characters because URLs, DOIs, and
   citation tags often contain underscores.

If a citekey does not match this regex, it can be substituted for a
tag that does, as defined in citation-tags.tsv.

https://github.com/greenelab/manubot-rootstock/issues/2#issuecomment-312153192

Prototyped at https://regex101.com/r/s3Asz3/4
"""
citekey_pattern = re.compile(r"(?<!\w)@([a-zA-Z0-9][\w:.#$%&\-+?<>~/]*[a-zA-Z0-9/])")


@functools.lru_cache(maxsize=5_000)
def standardize_citekey(citekey, warn_if_changed=False):
    """
    Standardize citation keys based on their source
    """
    source, identifier = citekey.split(":", 1)

    if source == "doi":
        if identifier.startswith("10/"):
            from manubot.cite.doi import expand_short_doi

            try:
                identifier = expand_short_doi(identifier)
            except Exception as error:
                # If DOI shortening fails, return the unshortened DOI.
                # DOI metadata lookup will eventually fail somewhere with
                # appropriate error handling, as opposed to here.
                logging.error(
                    f"Error in expand_short_doi for {identifier} "
                    f"due to a {error.__class__.__name__}:\n{error}"
                )
                logging.info(error, exc_info=True)
        identifier = identifier.lower()

    if source == "isbn":
        from isbnlib import to_isbn13

        identifier = to_isbn13(identifier)

    standard_citekey = f"{source}:{identifier}"
    if warn_if_changed and citekey != standard_citekey:
        logging.warning(
            f"standardize_citekey expected citekey to already be standardized.\n"
            f"Instead citekey was changed from {citekey!r} to {standard_citekey!r}"
        )
    return standard_citekey


regexes = {
    "pmid": re.compile(r"[1-9][0-9]{0,7}"),
    "pmcid": re.compile(r"PMC[0-9]+"),
    "doi": re.compile(r"10\.[0-9]{4,9}/\S+"),
    "shortdoi": re.compile(r"10/[a-zA-Z0-9]+"),
    "wikidata": re.compile(r"Q[0-9]+"),
}


def inspect_citekey(citekey):
    """
    Check citekeys adhere to expected formats. If an issue is detected a
    string describing the issue is returned. Otherwise returns None.
    """
    source, identifier = citekey.split(":", 1)

    if source == "pmid":
        # https://www.nlm.nih.gov/bsd/mms/medlineelements.html#pmid
        if identifier.startswith("PMC"):
            return (
                "PubMed Identifiers should start with digits rather than PMC. "
                f"Should {citekey!r} switch the citation source to 'pmcid'?"
            )
        elif not regexes["pmid"].fullmatch(identifier):
            return "PubMed Identifiers should be 1-8 digits with no leading zeros."

    if source == "pmcid":
        # https://www.nlm.nih.gov/bsd/mms/medlineelements.html#pmc
        if not identifier.startswith("PMC"):
            return "PubMed Central Identifiers must start with 'PMC'."
        elif not regexes["pmcid"].fullmatch(identifier):
            return (
                "Identifier does not conform to the PMCID regex. "
                "Double check the PMCID."
            )

    if source == "doi":
        if identifier.startswith("10."):
            # https://www.crossref.org/blog/dois-and-matching-regular-expressions/
            if not regexes["doi"].fullmatch(identifier):
                return (
                    "Identifier does not conform to the DOI regex. "
                    "Double check the DOI."
                )
        elif identifier.startswith("10/"):
            # shortDOI, see http://shortdoi.org
            if not regexes["shortdoi"].fullmatch(identifier):
                return (
                    "Identifier does not conform to the shortDOI regex. "
                    "Double check the shortDOI."
                )
        else:
            return "DOIs must start with '10.' (or '10/' for shortDOIs)."

    if source == "isbn":
        import isbnlib

        fail = isbnlib.notisbn(identifier, level="strict")
        if fail:
            return f"identifier violates the ISBN syntax according to isbnlib v{isbnlib.__version__}"

    if source == "wikidata":
        # https://www.wikidata.org/wiki/Wikidata:Identifiers
        if not identifier.startswith("Q"):
            return "Wikidata item IDs must start with 'Q'."
        elif not regexes["wikidata"].fullmatch(identifier):
            return (
                "Identifier does not conform to the Wikidata regex. "
                "Double check the entity ID."
            )

    return None


def is_valid_citekey(
    citekey, allow_tag=False, allow_raw=False, allow_pandoc_xnos=False
):
    """
    Return True if citekey is a properly formatted string. Return False if
    citekey is not a citation or is an invalid citation.

    In the case citekey is invalid, an error is logged. This
    function does not catch all invalid citekeys, but instead performs cursory
    checks, such as ensuring citekeys adhere to the expected formats. No calls to
    external resources are used by these checks, so they will not detect
    citekeys to non-existent identifiers unless those identifiers violate
    their source's syntax.

    allow_tag=False, allow_raw=False, and allow_pandoc_xnos=False enable
    allowing citekey sources that are valid for Manubot manuscripts, but
    likely not elsewhere. allow_tag=True enables citekey tags (e.g.
    tag:citation-tag). allow_raw=True enables raw citekeys (e.g.
    raw:manual-reference). allow_pandoc_xnos=True still returns False for
    pandoc-xnos references (e.g. fig:figure-id), but does not log an error.
    With the default of False for these arguments, valid sources are restricted
    to those for which manubot can retrieve metadata based only on the
    standalone citekey.
    """
    if not isinstance(citekey, str):
        logging.error(
            f"citekey should be type 'str' not "
            f"{type(citekey).__name__!r}: {citekey!r}"
        )
        return False
    if citekey.startswith("@"):
        logging.error(f"invalid citekey: {citekey!r}\nstarts with '@'")
        return False
    try:
        source, identifier = citekey.split(":", 1)
    except ValueError:
        logging.error(
            f"citekey not splittable via a single colon: {citekey}. "
            "Citekeys must be in the format of `source:identifier`."
        )
        return False

    if not source or not identifier:
        msg = f"invalid citekey: {citekey!r}\nblank source or identifier"
        logging.error(msg)
        return False

    if allow_pandoc_xnos:
        # Exempted non-citation sources used for pandoc-fignos,
        # pandoc-tablenos, and pandoc-eqnos
        pandoc_xnos_keys = {"fig", "tbl", "eq"}
        if source in pandoc_xnos_keys:
            return False
        if source.lower() in pandoc_xnos_keys:
            logging.error(
                f"pandoc-xnos reference types should be all lowercase.\n"
                f'Should {citekey!r} use {source.lower()!r} rather than "{source!r}"?'
            )
            return False

    # Check supported source type
    sources = set(citeproc_retrievers)
    if allow_raw:
        sources.add("raw")
    if allow_tag:
        sources.add("tag")
    if source not in sources:
        if source.lower() in sources:
            logging.error(
                f"citekey sources should be all lowercase.\n"
                f'Should {citekey} use "{source.lower()}" rather than "{source}"?'
            )
        else:
            logging.error(
                f"invalid citekey: {citekey!r}\n"
                f"Source {source!r} is not valid.\n"
                f'Valid citation sources are {{{", ".join(sorted(sources))}}}'
            )
        return False

    inspection = inspect_citekey(citekey)
    if inspection:
        logging.error(f"invalid {source} citekey: {citekey}\n{inspection}")
        return False

    return True


def shorten_citekey(standard_citekey):
    """
    Return a shortened citekey derived from the input citekey.
    The input citekey should be standardized prior to this function,
    since differences in the input citekey will result in different shortened citekeys.
    Short citekeys are generated by converting the input citekey to a 6 byte hash,
    and then converting this digest to a base62 ASCII str. Shortened
    citekeys consist of characters in the following ranges: 0-9, a-z and A-Z.
    """
    import hashlib
    import base62

    assert not standard_citekey.startswith("@")
    as_bytes = standard_citekey.encode()
    blake_hash = hashlib.blake2b(as_bytes, digest_size=6)
    digest = blake_hash.digest()
    short_citekey = base62.encodebytes(digest)
    return short_citekey


def citekey_to_csl_item(citekey, prune=True):
    """
    Generate a CSL Item (Python dictionary) for the input citekey.
    """
    from manubot.cite.csl_item import CSL_Item
    from manubot import __version__ as manubot_version

    citekey == standardize_citekey(citekey, warn_if_changed=True)
    source, identifier = citekey.split(":", 1)

    if source not in citeproc_retrievers:
        msg = f"Unsupported citation source {source!r} in {citekey!r}"
        raise ValueError(msg)
    citeproc_retriever = import_function(citeproc_retrievers[source])
    csl_item = citeproc_retriever(identifier)
    csl_item = CSL_Item(csl_item)

    note_text = f"This CSL JSON Item was automatically generated by Manubot v{manubot_version} using citation-by-identifier."
    note_dict = {"standard_id": citekey}
    csl_item.note_append_text(note_text)
    csl_item.note_append_dict(note_dict)

    short_citekey = shorten_citekey(citekey)
    csl_item.set_id(short_citekey)
    csl_item.clean(prune=prune)

    return csl_item


def infer_citekey_prefix(citekey):
    """
    Passthrough citekey if it has a valid citation key prefix. Otherwise,
    if the lowercase citekey prefix is valid, convert the prefix to lowercase.
    Otherwise, assume citekey is raw and prepend "raw:".
    """
    prefixes = [f"{x}:" for x in list(citeproc_retrievers) + ["raw"]]
    for prefix in prefixes:
        if citekey.startswith(prefix):
            return citekey
        if citekey.lower().startswith(prefix):
            return prefix + citekey[len(prefix) :]
    return f"raw:{citekey}"



import json
import logging
import subprocess

from manubot.pandoc.util import get_pandoc_info
from manubot.util import shlex_join


def load_bibliography(path=None, text=None, input_format=None):
    """
    Convert a bibliography to CSL JSON using `pandoc-citeproc --bib2json`.
    Accepts either a bibliography path or text (string). If supplying text,
    pandoc-citeproc will likely require input_format be specified.
    The CSL JSON is returned as Python objects.

    Parameters
    ----------
    path : str, pathlike, or None
        Path to a bibliography file. Extension is used by pandoc-citeproc to infer the
        format of the input.
    text : str or None
        Text representation of the bibligriophy, such as a JSON-formatted string.
        `input_format` should be specified if providing text input.
    input_format : str or None
        Manually specified input formatted that is supported by pandoc-citeproc:
        https://github.com/jgm/pandoc-citeproc/blob/master/man/pandoc-citeproc.1.md#options

    Returns
    -------
    csl_json : JSON-like object
        CSL JSON Data for the references encoded by the input bibliography.
    """
    use_text = path is None
    use_path = text is None
    if not (use_text ^ use_path):
        raise ValueError("load_bibliography: specify either path or text but not both.")
    if not get_pandoc_info()["pandoc-citeproc"]:
        logging.error(
            "pandoc-citeproc not found on system: manubot.pandoc.bibliography.load_bibliography returning empty CSL JSON"
        )
        return []
    args = ["pandoc-citeproc", "--bib2json"]
    if input_format:
        args.extend(["--format", input_format])
    run_kwargs = {}
    if use_path:
        args.append(str(path))
    if use_text:
        run_kwargs["input"] = text
    logging.info("call_pandoc subprocess args:\n>>> " + shlex_join(args))
    process = subprocess.run(
        args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        **run_kwargs,
    )
    logging.info(f"captured stderr:\n{process.stderr}")
    process.check_returncode()
    try:
        csl_json = json.loads(process.stdout)
    except Exception:
        logging.exception(f"Error parsing bib2json output as JSON:\n{process.stdout}")
        csl_json = []
    return csl_json



__all__ = [
    "citation_to_citeproc",
    "citekey_to_csl_item",
    "standardize_citation",
    "standardize_citekey",
]

from manubot.cite.citekey import citekey_to_csl_item, standardize_citekey


def citation_to_citeproc(*args, **kwargs):
    import warnings

    warnings.warn(
        "'citation_to_citeproc' has been renamed to 'citekey_to_csl_item'"
        " and will be removed in a future release.",
        category=FutureWarning,
    )
    return citekey_to_csl_item(*args, **kwargs)


def standardize_citation(*args, **kwargs):
    import warnings

    warnings.warn(
        "'standardize_citation' has been renamed to 'standardize_citekey'"
        " and will be removed in a future release.",
        category=FutureWarning,
    )
    return standardize_citekey(*args, **kwargs)



import json
import logging
import pathlib
import subprocess
import sys

from manubot.cite.citekey import (
    citekey_to_csl_item,
    standardize_citekey,
    is_valid_citekey,
)
from manubot.pandoc.util import get_pandoc_info
from manubot.util import shlex_join


# For manubot cite, infer --format from --output filename extensions
extension_to_format = {
    ".txt": "plain",
    ".md": "markdown",
    ".docx": "docx",
    ".html": "html",
    ".xml": "jats",
}


def call_pandoc(metadata, path, format="plain"):
    """
    path is the path to write to.
    """
    _exit_without_pandoc()
    info = get_pandoc_info()
    _check_pandoc_version(info, metadata, format)
    metadata_block = "---\n{yaml}\n...\n".format(
        yaml=json.dumps(metadata, ensure_ascii=False, indent=2)
    )
    args = [
        "pandoc",
        "--filter",
        "pandoc-citeproc",
        "--output",
        str(path) if path else "-",
    ]
    if format == "markdown":
        args.extend(["--to", "markdown_strict", "--wrap", "none"])
    elif format == "jats":
        args.extend(["--to", "jats", "--standalone"])
    elif format == "docx":
        args.extend(["--to", "docx"])
    elif format == "html":
        args.extend(["--to", "html"])
    elif format == "plain":
        args.extend(["--to", "plain", "--wrap", "none"])
        if info["pandoc version"] >= (2,):
            # Do not use ALL_CAPS for bold & underscores for italics
            # https://github.com/jgm/pandoc/issues/4834#issuecomment-412972008
            filter_path = (
                pathlib.Path(__file__)
                .joinpath("..", "plain-pandoc-filter.lua")
                .resolve()
            )
            assert filter_path.exists()
            args.extend(["--lua-filter", str(filter_path)])
    logging.info("call_pandoc subprocess args:\n" + shlex_join(args))
    process = subprocess.run(
        args=args,
        input=metadata_block.encode(),
        stdout=subprocess.PIPE if path else sys.stdout,
        stderr=sys.stderr,
    )
    process.check_returncode()


def cli_cite(args):
    """
    Main function for the manubot cite command-line interface.

    Does not allow user to directly specify Pandoc's --to argument, due to
    inconsistent citaiton rendering by output format. See
    https://github.com/jgm/pandoc/issues/4834
    """
    # generate CSL JSON data
    csl_list = list()
    for citekey in args.citekeys:
        try:
            if not is_valid_citekey(citekey):
                continue
            citekey = standardize_citekey(citekey)
            csl_item = citekey_to_csl_item(citekey, prune=args.prune_csl)
            csl_list.append(csl_item)
        except Exception as error:
            logging.error(
                f"citekey_to_csl_item for {citekey!r} failed "
                f"due to a {error.__class__.__name__}:\n{error}"
            )
            logging.info(error, exc_info=True)

    # output CSL JSON data, if --render is False
    if not args.render:
        write_file = (
            args.output.open("w", encoding="utf-8") if args.output else sys.stdout
        )
        with write_file:
            json.dump(csl_list, write_file, ensure_ascii=False, indent=2)
            write_file.write("\n")
        return

    # use Pandoc to render references
    if not args.format and args.output:
        vars(args)["format"] = extension_to_format.get(args.output.suffix)
    if not args.format:
        vars(args)["format"] = "plain"
    pandoc_metadata = {"nocite": "@*", "csl": args.csl, "references": csl_list}
    call_pandoc(metadata=pandoc_metadata, path=args.output, format=args.format)


def _exit_without_pandoc():
    """
    Given info from get_pandoc_info, exit Python if Pandoc is not available.
    """
    info = get_pandoc_info()
    for command in "pandoc", "pandoc-citeproc":
        if not info[command]:
            logging.critical(
                f'"{command}" not found on system. ' f"Check that Pandoc is installed."
            )
            raise SystemExit(1)


def _check_pandoc_version(info, metadata, format):
    """
    Given info from get_pandoc_info, check that Pandoc's version is sufficient
    to perform the citation rendering command specified by metadata and format.
    Please add additional minimum version information to this function, as its
    discovered.
    """
    issues = list()
    if format == "jats" and info["pandoc version"] < (2,):
        issues.append("--jats requires pandoc >= v2.0.")
    # --csl=URL did not work in https://travis-ci.org/greenelab/manubot/builds/417314743#L796,
    # but exact version where this fails unknown
    # if metadata.get('csl', '').startswith('http') and pandoc_version < (2,):
    #     issues.append('--csl=URL requires pandoc >= v2.0.')
    issues = "\n".join(issues)
    if issues:
        logging.critical(f"issues with pandoc version detected:\n{issues}")



import logging
import shutil
import subprocess
import functools


@functools.lru_cache()
def get_pandoc_info():
    """
    Return path and version information for the system's pandoc and
    pandoc-citeproc commands. When Pandoc is installed,
    the output will look like:
    ```python
    {
        'pandoc': True,
        'pandoc path': '/PATH_TO_EXECUTABLES/pandoc',
        'pandoc version': (2, 5),
        'pandoc-citeproc': True,
        'pandoc-citeproc path': '/PATH_TO_EXECUTABLES/pandoc-citeproc',
        'pandoc-citeproc version': (0, 15),
    }
    ```

    If the executables are missing, the output will be like:
    ```python
    {
        'pandoc': False,
        'pandoc-citeproc': False,
    }
    ```
    """
    stats = dict()
    for command in "pandoc", "pandoc-citeproc":
        path = shutil.which(command)
        stats[command] = bool(path)
        if not path:
            continue
        version = subprocess.check_output(
            args=[command, "--version"], universal_newlines=True
        )
        logging.debug(version)
        version, *discard = version.splitlines()
        discard, version = version.strip().split()
        from packaging.version import parse as parse_version

        version = parse_version(version).release
        stats[f"{command} version"] = version
        stats[f"{command} path"] = path
    logging.info("\n".join(f"{k}: {v}" for k, v in stats.items()))
    return stats


def get_pandoc_version() -> (int, int, int):
    """
    Return pandoc version as tuple of major and minor
    version numbers, for example: (2, 7, 2)
    """
    return get_pandoc_info()["pandoc version"]



import copy

import pytest


from ..metadata import get_header_includes, get_thumbnail_url, get_manuscript_urls
from ..ci import get_continuous_integration_parameters


def test_get_header_includes_description_abstract():
    # test that abstract and description set different fields if both supplied
    variables = {
        "pandoc": {"abstract": "value for abstract"},
        "manubot": {"description": "value for description"},
    }
    header_includes = get_header_includes(copy.deepcopy(variables))
    assert (
        '<meta name="citation_abstract" content="value for abstract" />'
        in header_includes
    )
    assert (
        '<meta name="description" content="value for description" />' in header_includes
    )
    # test that abstract is used for description if description is not defined
    del variables["manubot"]["description"]
    header_includes = get_header_includes(copy.deepcopy(variables))
    assert (
        '<meta name="citation_abstract" content="value for abstract" />'
        in header_includes
    )
    assert '<meta name="description" content="value for abstract" />' in header_includes
    # test that if neither abstract nor description is defined, neither are inserted
    del variables["pandoc"]["abstract"]
    header_includes = get_header_includes(copy.deepcopy(variables))
    assert 'meta name="citation_abstract"' not in header_includes
    assert 'meta name="description"' not in header_includes


ci_params = get_continuous_integration_parameters() or {}
local_only = pytest.mark.skipif(
    ci_params, reason="skipping on local build since test assumes supported CI behavior"
)
ci_only = pytest.mark.skipif(
    not ci_params, reason="skipping on CI build since test assumes local behavior"
)

repo_raw_url_template = (
    f"https://github.com/{ci_params.get('repo_slug', '')}"
    f"/raw/{ci_params.get('triggering_commit', '')}/"
)
example_thumbnail_url = (
    repo_raw_url_template + "manubot/process/tests/manuscripts/example/thumbnail.png"
)


@pytest.mark.parametrize(
    ("thumbnail", "expected"),
    [
        pytest.param(None, None, id="None-on-local", marks=local_only),
        pytest.param("", None, id="empty-on-local", marks=local_only),
        pytest.param(None, example_thumbnail_url, id="None-on-ci", marks=ci_only),
        pytest.param("", example_thumbnail_url, id="empty-on-ci", marks=ci_only),
        pytest.param(
            "path/to/thumbnail.png", None, id="local-path-on-local", marks=local_only
        ),
        pytest.param(
            "path/to/thumbnail.png",
            repo_raw_url_template + "path/to/thumbnail.png",
            id="local-path-on-ci",
            marks=ci_only,
        ),
        pytest.param(
            "http://example.com/thumbnail.png",
            "http://example.com/thumbnail.png",
            id="url-http",
        ),
        pytest.param(
            "https://example.com/thumbnail.png",
            "https://example.com/thumbnail.png",
            id="url-https",
        ),
    ],
)
def test_get_thumbnail_url(thumbnail, expected):
    thumbnail_url = get_thumbnail_url(thumbnail)
    assert expected == thumbnail_url


@pytest.mark.parametrize(
    ("html_url", "expected"),
    [
        pytest.param(None, {}, id="html_url-none-local", marks=local_only),
        pytest.param(
            "https://example.com/manuscript/",
            {
                "html_url": "https://example.com/manuscript/",
                "pdf_url": "https://example.com/manuscript/manuscript.pdf",
            },
            id="html_url-set-local",
            marks=local_only,
        ),
        pytest.param(
            None,
            {
                "html_url": "https://manubot.github.io/manubot/",
                "pdf_url": "https://manubot.github.io/manubot/manuscript.pdf",
                "html_url_versioned": f"https://manubot.github.io/manubot/v/{ci_params.get('commit')}/",
                "pdf_url_versioned": f"https://manubot.github.io/manubot/v/{ci_params.get('commit')}/manuscript.pdf",
            },
            id="html_url-none-ci",
            marks=ci_only,
        ),
        pytest.param(
            "https://example.com/manuscript/",
            {
                "html_url": "https://example.com/manuscript/",
                "pdf_url": "https://example.com/manuscript/manuscript.pdf",
                "html_url_versioned": f"https://example.com/manuscript/v/{ci_params.get('commit')}/",
                "pdf_url_versioned": f"https://example.com/manuscript/v/{ci_params.get('commit')}/manuscript.pdf",
            },
            id="html_url-set-ci",
            marks=ci_only,
        ),
    ],
)
def test_get_manuscript_urls(html_url, expected):
    urls = get_manuscript_urls(html_url)
    assert urls == expected
